# pyright: reportUnusedExpression=false
import os
import re
import json
import time
import textwrap
import ast
import uuid

import snowflake.snowpark
from .. import std
from collections import defaultdict
from railib.api import TransactionAsyncResponse, poll_with_specified_overhead, _parse_metadata_proto
import requests
import snowflake.connector
import pyarrow as pa

from snowflake.snowpark import Session
from . import result_helpers
from .. import debugging
from typing import Any, Dict, Iterable, Tuple, List, cast

from pandas import DataFrame

from ..tools.cli_controls import Spinner
from ..clients.types import AvailableModel, Import, ImportSource, ImportSourceTable
from ..clients.config import Config
from ..clients.client import Client, ExportParams, ResourceProvider
from ..clients.util import escape_for_f_string, escape_for_sproc, scrub_exception
from .. import dsl, rel, metamodel as m
from ..errors import EngineNotFoundException, EnginePending, Errors, RAIException, SnowflakeAppMissingException, SnowflakeChangeTrackingNotEnabledException, SnowflakeImportMissingException, SnowflakeInvalidSource, SnowflakeRaiAppNotStarted
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime

#--------------------------------------------------
# Constants
#--------------------------------------------------

VALID_POOL_STATUS = ["ACTIVE", "IDLE", "SUSPENDED"]
# transaction list and get return different fields (duration vs timings)
LIST_TXN_SQL_FIELDS = ["id", "database_name", "state", "abort_reason", "read_only", "created_by", "created_on", "finished_at", "duration"]
GET_TXN_SQL_FIELDS = ["id", "database_name", "engine_name", "state", "abort_reason", "read_only",
                      "created_by", "created_on", "finished_at", "timings"]
IMPORT_STREAM_FIELDS = ["ID", "CREATED_AT", "CREATED_BY", "STATUS", "REFERENCE_NAME", "REFERENCE_ALIAS", "FQ_OBJECT_NAME", "RAI_DATABASE",
                        "RAI_RELATION", "DATA_SYNC_STATUS", "PENDING_BATCHES_COUNT", "NEXT_BATCH_STATUS", "NEXT_BATCH_UNLOADED_TIMESTAMP",
                        "NEXT_BATCH_DETAILS", "LAST_BATCH_DETAILS", "LAST_BATCH_UNLOADED_TIMESTAMP", "CDC_STATUS"]
VALID_ENGINE_STATES = ["READY", "PENDING"]
ENGINE_SIZES = ["HighMem|S", "HighMem|M"]
COMPUTE_POOL_MAP = {
    "HIGHMEM_X64_S": "HighMem|S",
    "HIGHMEM_X64_M": "HighMem|M",
    "HIGHMEM_S": "HighMem|S",
    "HIGHMEM_M": "HighMem|M",
    "HighMem|S": "HighMem|S",
    "HighMem|M": "HighMem|M",
}
VALID_IMPORT_STATES = ["PENDING", "PROCESSING", "QUARANTINED", "LOADED"]
RAI_APP_NAME = "RELATIONALAI"

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def process_jinja_template(template: str, indent_spaces = 0, **substitutions) -> str:
    lines = template.split('\n')
    processed_lines = []

    for line in lines:
        processed_line = re.sub(r'{{\s*(\w+)\s*}}', lambda m: str(substitutions[m.group(1)]), line)
        processed_lines.append((indent_spaces * " ") + processed_line)

    return '\n'.join(processed_lines)

def type_to_sql(type) -> str:
    if type is str:
        return "VARCHAR"
    if type is int:
        return "NUMBER"
    if type is float:
        return "FLOAT"
    if type is bool:
        return "BOOLEAN"
    if type is dict:
        return "VARIANT"
    if type is list:
        return "ARRAY"
    if type is bytes:
        return "BINARY"
    if isinstance(type, dsl.Type):
        return "NUMBER"
    raise ValueError(f"Unknown type {type}")

def type_to_snowpark(type) -> str:
    if type is str:
        return "StringType"
    if type is int:
        return "IntegerType"
    if type is float:
        return "FloatType"
    if type is bool:
        return "BooleanType"
    if type is dict:
        return "MapType"
    if type is list:
        return "ArrayType"
    if type is bytes:
        return "BinaryType"
    if isinstance(type, dsl.Type):
        return "IntegerType"
    raise ValueError(f"Unknown type {type}")

#--------------------------------------------------
# Resources
#--------------------------------------------------

APP_NAME = "___RAI_APP___"

class Resources(ResourceProvider):
    def __init__(
        self,
        profile: str | None = None,
        config: Config | None = None,
        connection: Session | None = None,
    ):
        super().__init__(profile, config=config)
        self._session = connection
        self._pending_transactions: list[str] = []
        self._ns_cache = {}

    def _exec(self, code:str, params:List[Any]|None = None, raw=False) -> Any:
        if not self._session:
            connection_parameters = {
                "account": self.config.get('account'),
                "user": self.config.get('user'),
                "password": self.config.get('password'),
                "role": self.config.get('role', ""),
                "warehouse":  self.config.get('warehouse', ""),
            }
            self._session = Session.builder.configs(connection_parameters).create()
        try:
            if params is not None and not isinstance(params, list):
                params = [params]
            sess_results = self._session.sql(
                code.replace(APP_NAME, self.get_app_name()),
                params
            )
            if raw:
                return sess_results
            return sess_results.collect()
        except Exception as e:
            orig_message = str(e).lower()
            rai_app = self.config.get("rai_app_name", "")
            if re.search(f"database '{rai_app}' does not exist or not authorized.".lower(), orig_message):
                print("\n")
                exception = SnowflakeAppMissingException(rai_app)
                raise exception from None
            elif re.search(r"javascript execution error", orig_message):
                match = re.search(r"\"message\":\"(.*)\"", orig_message)
                if match:
                    message = match.group(1)
                    if "engine was deleted" in message or "engine not found" in message:
                        exception = EngineNotFoundException(self.config.get('engine', "Unknown"), message)
                        raise exception from None
                    if "engine is in pending" in message:
                        raise EnginePending(self.config.get('engine', "Unknown"))
                    else:
                        raise RAIException(message) from None

            if re.search(r"the relationalai service has not been started.", orig_message):
                app_name = self.config.get("rai_app_name", "")
                sql = "call app.start_service(compute_pool_name, warehouse_name)"
                raise SnowflakeRaiAppNotStarted(app_name, sql)
            raise  RAIException(str(e))


    def reset(self):
        self._session = None

    #--------------------------------------------------
    # Databases
    #--------------------------------------------------

    def get_database(self, name: str):
        results = self._exec(
            f"SELECT * from {APP_NAME}.api.DATABASES WHERE name like '{name}';"
        )
        if not results:
            return None
        db = results[0]
        if not db:
            return None
        return {
            "id": db["ID"],
            "name": db["NAME"],
            "created_by": db["CREATED_BY"],
            "created_on": db["CREATED_ON"],
            "deleted_by": db["DELETED_BY"],
            "deleted_on": db["DELETED_ON"],
            "state": db["STATE"],
        }

    #--------------------------------------------------
    # Engines
    #--------------------------------------------------
    def get_engine_sizes(self):
        return ENGINE_SIZES

    def list_engines(self, state: str | None = None):
        where_clause = f"WHERE STATUS = '{state.upper()}'" if state else ""
        statement = f"SELECT NAME, ID, SIZE, STATUS, CREATED_BY, CREATED_ON, UPDATED_ON FROM {APP_NAME}.api.engines {where_clause}"
        results = self._exec(statement)
        if not results:
            return []
        return [
            {
                "name": row["NAME"],
                "id": row["ID"],
                "size": row["SIZE"],
                "state": row["STATUS"], # callers are expecting 'state'
                "created_by": row["CREATED_BY"],
                "created_on": row["CREATED_ON"],
                "updated_on": row["UPDATED_ON"],
            }
            for row in results
        ]

    def get_engine(self, name: str):
        results = self._exec(
            f"SELECT NAME, ID, SIZE, STATUS, CREATED_BY, CREATED_ON, UPDATED_ON FROM {APP_NAME}.api.engines WHERE NAME='{name}'"
        )
        if not results:
            return None
        engine = results[0]
        if not engine:
            return None
        return {
            "name": engine["NAME"],
            "id": engine["ID"],
            "size": engine["SIZE"],
            "state": engine["STATUS"], # callers are expecting 'state'
            "created_by": engine["CREATED_BY"],
            "created_on": engine["CREATED_ON"],
            "updated_on": engine["UPDATED_ON"],
        }

    def is_valid_engine_state(self, state:str):
        return state in VALID_ENGINE_STATES

    def create_engine(self, name:str, size:str, pool:str = ""):
        if not pool:
            raise ValueError("Pool is required")
        try:
            with debugging.span("create_engine", name=name, size=size, pool=pool):
                self._exec(f"call {APP_NAME}.api.create_engine('{name}', '{pool}', '{size}');")
        except Exception as e:
            raise Exception(f"Failed to create engine {name} using pool {pool} and size {size}: {e}") from e

    def delete_engine(self, name:str, force_delete:bool = False):
        self._exec(f"call {APP_NAME}.api.delete_engine('{name}', {force_delete});")

    def suspend_engine(self, name:str):
        raise Exception("Not implemented")

    def resume_engine(self, name:str):
        raise Exception("Not implemented")

    #--------------------------------------------------
    # Graphs
    #--------------------------------------------------

    def list_graphs(self) -> List[AvailableModel]:
        with debugging.span("list_models"):
            query = textwrap.dedent(f"""
                    SELECT NAME, ID, CREATED_BY, CREATED_ON, STATE, DELETED_BY, DELETED_ON
                    FROM {APP_NAME}.api.databases
                    WHERE state <> 'DELETED'
                    """)
            results = self._exec(query)
            if not results:
                return []
            return [
                {
                    "name": row["NAME"],
                    "id": row["ID"],
                    "created_by": row["CREATED_BY"],
                    "created_on": row["CREATED_ON"],
                    "state": row["STATE"],
                    "deleted_by": row["DELETED_BY"],
                    "deleted_on": row["DELETED_ON"],
                }
                for row in results
            ]

    def get_graph(self, name: str):
        with debugging.span("get_model", name=name):
            results = self._exec(
                textwrap.dedent(f"""
                SELECT ID, NAME, CREATED_BY, CREATED_ON, DELETED_BY, DELETED_ON, STATE
                FROM {APP_NAME}.api.databases
                WHERE NAME='{name}' AND STATE <> 'DELETED'
                """)
            )
            if not results:
                return None
            return results[0]

    def create_graph(self, name: str):
        with debugging.span("create_model", name=name):
            self._exec(f"call {APP_NAME}.api.create_database('{name}');")

    def delete_graph(self, name:str):
        with debugging.span("delete_model", name=name):
            self._exec(f"call {APP_NAME}.api.delete_database('{name}');")

    def clone_graph(self, target_name:str, source_name:str, nowait_durable:bool = True):
        with debugging.span("clone_model", target_name=target_name, source_name=source_name):
            # not a mistake: the clone_database argument order is indeed target then source:
            self._exec(f"call {APP_NAME}.api.clone_database('{target_name}', '{source_name}', {nowait_durable});")

    #--------------------------------------------------
    # Models
    #--------------------------------------------------

    def list_models(self, database: str, engine: str):
        pass

    def create_models(self, database: str, engine: str, models:List[Tuple[str, str]]) -> List[Any]:
        rel_code = self.create_models_code(models)
        self.exec_raw(database, engine, rel_code, readonly=False)
        # TODO: handle SPCS errors once they're figured out
        return []

    def delete_model(self, database:str, engine:str, name:str):
        self.exec_raw(database, engine, f"def delete[:rel, :catalog, :model, \"{name}\"]: rel[:catalog, :model, \"{name}\"]", readonly=False)

    def create_models_code(self, models:List[Tuple[str, str]]) -> str:
        lines = []
        for (name, code) in models:
            name = name.replace("\"", "\\\"")
            assert "\"\"\"\"\"\"\"" not in code, "Code literals must use fewer than 7 quotes."

            lines.append(textwrap.dedent(f"""
            def delete[:rel, :catalog, :model, "{name}"]: rel[:catalog, :model, "{name}"]
            def insert[:rel, :catalog, :model, "{name}"]: raw\"\"\"\"\"\"\"
            """) + code + "\n\"\"\"\"\"\"\"")
        rel_code = "\n\n".join(lines)
        return rel_code

    #--------------------------------------------------
    # Exports
    #--------------------------------------------------

    def list_exports(self, database: str, engine: str):
        return []

    def get_export_code(self, params: ExportParams):
        sql_inputs = ", ".join([f"{name} {type_to_sql(type)}" for (name, _, type) in params.inputs])
        input_names = [name for (name, *_) in params.inputs]
        sql_out = ", ".join([f"{name} {type_to_sql(type)}" for (name, type) in params.out_fields])
        py_outs = ", ".join([f"StructField(\"{name}\", {type_to_snowpark(type)}())" for (name, type) in params.out_fields])
        py_inputs = ", ".join([name for (name, *_) in params.inputs])
        safe_rel = escape_for_f_string(params.code).strip()
        clean_inputs = []
        for (name, var, type) in params.inputs:
            if type is str:
                clean_inputs.append(f"{name} = '\"' + escape({name}) + '\"'")
            # Replace `var` with `name` and keep the following non-word character unchanged
            pattern = re.compile(re.escape(var) + r'(\W)')
            safe_rel = re.sub(pattern, rf"{{{name}}}\1", safe_rel)
        if py_inputs:
            py_inputs = f", {py_inputs}"
        clean_inputs = ("\n" + " "*20).join(clean_inputs)
        template_path = os.path.join(os.path.dirname(__file__), "export_procedure.py.jinja")
        with open(template_path) as f:
            template = f.read()
        def quote(s: str, f = False) -> str:
            return '"' + s + '"' if not f else 'f"' + s + '"'
        python_code = process_jinja_template(
            template,
            indent_spaces=12,
            func_name=quote(params.func_name),
            database=quote(params.source_database),
            persistent_database=quote(params.database),
            engine=quote(params.engine),
            rel_code=quote(safe_rel, f=True),
            APP_NAME=quote(APP_NAME),
            input_names=input_names,
            outputs=sql_out,
            clean_inputs=clean_inputs,
            py_inputs=py_inputs,
            py_outs=py_outs,
        ).strip()
        sql_code = textwrap.dedent(f"""
            CREATE OR REPLACE PROCEDURE {params.func_name}({sql_inputs}{sql_inputs and ',' or ''} engine STRING DEFAULT NULL)
            RETURNS TABLE({sql_out})
            LANGUAGE PYTHON
            RUNTIME_VERSION = '3.8'
            PACKAGES = ('snowflake-snowpark-python')
            HANDLER = 'handle'
            EXECUTE AS CALLER
            AS
            $$
            {escape_for_sproc(python_code)}
            $$;
        """)
        python_code = textwrap.dedent(" " * 12 + python_code)
        # This check helps catch invalid code early and for dry runs:
        try:
            ast.parse(python_code)
        except SyntaxError:
            raise ValueError(f"Internal error: invalid Python code generated:\n{python_code}")
        return sql_code


    def create_export(self, params: ExportParams):
        self.delete_on_exit = False
        sql_code = self.get_export_code(params)
        if not params.dry_run:
            start = time.perf_counter()
            self._exec(sql_code)
            debugging.time("export", time.perf_counter() - start, DataFrame(), code=sql_code.replace(APP_NAME, self.get_app_name()))


    def create_export_table(self, database: str, engine: str, table: str, relation: str, columns: Dict[str, str], code: str, refresh: str|None=None):
        print("Snowflake doesn't support creating export tables yet. Try creating the table manually first.")
        pass

    def delete_export(self, model: str, engine: str, export: str):
        pass

    #--------------------------------------------------
    # Imports
    #--------------------------------------------------

    def is_valid_import_state(self, state:str):
        return state in VALID_IMPORT_STATES

    def imports_to_dicts(self, results):
        parsed_results = [
            {field.lower(): row[field] for field in IMPORT_STREAM_FIELDS}
            for row in results
        ]
        return parsed_results

    def change_stream_status(self, stream_id: str, model:str, suspend: bool):
        if stream_id and model:
            if suspend:
                self._exec(f"CALL {APP_NAME}.api.suspend_data_stream('{stream_id}', '{model}');")
            else:
                self._exec(f"CALL {APP_NAME}.api.resume_data_stream('{stream_id}', '{model}');")

    def change_imports_status(self, suspend: bool):
        if suspend:
            self._exec(f"CALL {APP_NAME}.app.suspend_cdc();")
        else:
            self._exec(f"CALL {APP_NAME}.app.resume_cdc();")

    def get_imports_status(self):
        results = self._exec(f"CALL {APP_NAME}.app.cdc_status();")
        if results:
            for result in results:
                engine = result['CDC_ENGINE_NAME']
                status = result['CDC_TASK_STATUS']
                info = result['CDC_TASK_INFO']
                return {"engine": engine, "status": status, "info": info}
        return {}

    def set_imports_engine(self, engine:str):
        try:
            self._exec(f"CALL {APP_NAME}.app.setup_cdc('{engine}');")
        except Exception as e:
            if "engine not found" in str(e):
                raise ValueError(f"Engine {engine} not found. Please create the engine first.") from e

    def list_imports(
        self,
        id:str|None = None,
        name:str|None = None,
        model:str|None = None,
        status:str|None = None,
        creator:str|None = None,
    ) -> list[Import]:
        where = []
        if id and isinstance(id, str):
            where.append(f"LOWER(ID) = '{id.lower()}'")
        if name and isinstance(name, str):
            where.append(f"LOWER(FQ_OBJECT_NAME) = '{name.lower()}'")
        if model and isinstance(model, str):
            where.append(f"LOWER(RAI_DATABASE) = '{model.lower()}'")
        if creator and isinstance(creator, str):
            where.append(f"LOWER(CREATED_BY) = '{creator.lower()}'")
        if status and isinstance(status, str):
            where.append(f"LOWER(batch_status) = '{status.lower()}'")
        where_clause = " AND ".join(where)

        # this is roughly copied from the native app code because we don't have a way to
        # get the status of multiple streams at once and doing them individually is way
        # too slow
        statement = f"""
        select ID, RAI_DATABASE, FQ_OBJECT_NAME, CREATED_AT, CREATED_BY, nextBatch.status as batch_status, processing_errors, nextBatch.batches,
            from {APP_NAME}.api.data_streams as ds
            left JOIN (
            select
                data_stream_id,
                min_by(status, unloaded) as status,
                min_by(batch_details, unloaded) as batch_details,
                min(processing_details:processingErrors) as processing_errors,
                min(unloaded) as unloaded,
                count(*) as batches
            from {APP_NAME}.api.data_stream_batches
            group by data_stream_id
            ) nextBatch
            ON ds.id = nextBatch.data_stream_id
            {f"where {where_clause}" if where_clause else ""};
        """

        results = self._exec(statement)
        items = []
        if results:
            for stream in results:
                (id, db, name, created_at, created_by, status, processing_errors, batches) = stream
                if status and isinstance(status, str):
                    status = status.upper()
                if processing_errors:
                    if status in ["QUARANTINED", "PENDING"]:
                        start = processing_errors.rfind("Error")
                        if start != -1:
                            processing_errors = processing_errors[start:-1]
                    else:
                        processing_errors = None
                items.append(cast(Import, {
                    "id": id,
                    "model": db,
                    "name": name,
                    "created": created_at,
                    "creator": created_by,
                    "status": status.upper() if status else None,
                    "errors": processing_errors if processing_errors != "[]" else None,
                    "batches": f"{batches}" if batches else "",
                }))
        return items

    def poll_imports(self, sources:List[str], model:str):
        def check_imports():
            imports = [import_ for import_ in self.list_imports(model=model) if import_["name"] in sources]
            all_completed = all(import_["status"] == "LOADED" for import_ in imports)
            failed_imports = [import_["name"] for import_ in imports if import_["status"] == "QUARANTINED"]
            if all_completed:
                return True
            if failed_imports:
                raise RAIException("Imports failed:" + ", ".join(failed_imports)) from None
        poll_with_specified_overhead(check_imports, overhead_rate=0.2, max_delay=10)

    def get_import_stream(self, stream_id:str, model:str):
        results = self._exec(f"CALL {APP_NAME}.api.get_data_stream('{stream_id}', '{model}');")
        if not results:
            return None
        return self.imports_to_dicts(results)

    def create_import_stream(self, source:ImportSource, model:str, rate = 1, options: dict|None = None):
        assert isinstance(source, ImportSourceTable), "Snowflake integration only supports loading from SF Tables. Try loading your data as a table via the Snowflake interface first."
        object = source.fqn

        if object.lower() in [x["name"].lower() for x in self.list_imports(model=model)]:
            return

        query = f"SHOW OBJECTS LIKE '{source.table}' IN {source.database}.{source.schema}"
        info = self._exec(query)
        if not info:
            raise ValueError(f"Object {source.table} not found in {source.database}.{source.schema}")
        else:
            data = info[0]
            if not data:
                raise ValueError(f"Object {source.table} not found in {source.database}.{source.schema}")
            # (time, name, db_name, schema_name, kind, *rest)
            kind = data["kind"]

        try:
            self._exec(f"ALTER {kind} {object} SET CHANGE_TRACKING = TRUE;")
        except Exception:
            pass

        command = f"""call {APP_NAME}.api.create_data_stream(
            {APP_NAME}.api.object_reference('{kind}', '{object}'),
            '{model}',
            '{object.replace('.', '_').lower()}');"""
        try:
            self._exec(command)
        except Exception as e:
            if "ensure that CHANGE_TRACKING is enabled on the source object" in str(e):
                print("\n")
                exception = SnowflakeChangeTrackingNotEnabledException(object, f"ALTER TABLE {object} SET CHANGE_TRACKING = TRUE;")
                raise exception from None
            raise e

        return

    def create_import_snapshot(self, source:ImportSource, model:str, _: dict|None = None):
        raise Exception("Snowflake integration doesn't support snapshot imports yet")

    def delete_import(self, import_name:str, model:str):
        self._exec(f"""call {APP_NAME}.api.delete_data_stream(
            '{import_name}',
            '{model}'
        );""")
        return

    #--------------------------------------------------
    # Exec Async
    #--------------------------------------------------

    def _check_exec_async_status(self, txn_id: str):
        """Check whether the given transaction has completed."""

        with debugging.span("check_status"):
            response = self._exec(f"CALL {APP_NAME}.api.get_transaction('{txn_id}');")
            assert response, f"No results from get_transaction('{txn_id}')"

        response_row = next(iter(response))
        status: str = response_row['STATE']

        # remove the transaction from the pending list if it's completed or aborted
        if status in ["COMPLETED", "ABORTED"]:
            if txn_id in self._pending_transactions:
                self._pending_transactions.remove(txn_id)

        if status == "ABORTED":
            # see https://github.com/RelationalAI/snowflake-native-app/blob/d2f81bffca474fcec665101b7e0bb47fce2789a5/lib/integration/rai.sql#L956
            abort_reason = response_row['ABORT_REASON']
            raise Exception(
                f"Transaction aborted while waiting for results '{txn_id}' with reason '{abort_reason}'"
            )

        return status == "COMPLETED"

    def _list_exec_async_artifacts(self, txn_id: str) -> Dict[str, str]:
        """Grab the list of artifacts produced in the transaction and the URLs to retrieve their contents."""
        with debugging.span("list_results"):
            response = self._exec(
                f"CALL {APP_NAME}.api.list_transaction_outputs('{txn_id}');"
            )
            assert response, f"No results from list_transaction_outputs('{txn_id}')"
            return {row["FILE_NAME"]: row["PRESIGNED_URL"] for row in response}

    def _fetch_exec_async_artifacts(
        self, artifact_urls: Dict[str, str]
    ) -> Dict[str, Any]:
        """Grab the contents of the given artifacts from SF in parallel using threads."""
        contents = {}

        with requests.Session() as session:
            # Define _fetch_data inside the session context to close over 'session'
            def _fetch_data(name_url):
                name, url = name_url
                try:
                    response = session.get(url)
                    response.raise_for_status()  # Throw if something goes wrong
                    if name.endswith(".json"):
                        return name, response.json()
                    else:
                        return name, response.content
                except requests.RequestException as e:
                    raise scrub_exception(e)

            # Create a list of tuples for the map function
            name_url_pairs = list(artifact_urls.items())

            # Create ThreadPoolExecutor to handle requests in parallel
            with ThreadPoolExecutor(max_workers=5) as executor:
                # Map fetch across all URLs using the executor
                results = executor.map(_fetch_data, name_url_pairs)

                # Populate contents dictionary
                for name, data in results:
                    contents[name] = data

        return contents

    def _parse_exec_async_results(self, arrow_files: List[Tuple[str, bytes]]):
        """Mimics the logic in _parse_arrow_results of railib/api.py#L303 without requiring a wrapping multipart form."""
        results = []

        for file_name, file_content in arrow_files:
            with pa.ipc.open_stream(file_content) as reader:
                schema = reader.schema
                batches = [batch for batch in reader]
                table = pa.Table.from_batches(batches=batches, schema=schema)
                results.append({"relationId": file_name, "table": table})

        return results

    def _download_results(self, artifact_urls, txn_id, state: str) -> TransactionAsyncResponse:
        with debugging.span("download_results"):
            # Fetch artifacts
            artifacts = self._fetch_exec_async_artifacts(artifact_urls)

            # Parse metadata from the artifacts
            meta = _parse_metadata_proto(artifacts["metadata.proto"])
            meta_json = artifacts["metadata.json"]

            # Use the metadata to map arrow files to the relations they contain
            arrow_files_to_relations = {}
            for ix, relation in enumerate(meta.relations):
                arrow_file = relation.file_name
                relation_id = meta_json[ix]["relationId"]
                arrow_files_to_relations[arrow_file] = relation_id

            # Hydrate the arrow files into tables
            results = self._parse_exec_async_results([
                (arrow_files_to_relations[name], content) for name, content in artifacts.items() if name.endswith(".arrow")
            ])

            # Create and return the response
            rsp = TransactionAsyncResponse()
            rsp.transaction = {"id": txn_id, "state": state, "response_format_version": None}
            rsp.metadata = meta
            rsp.problems = artifacts.get("problems.json")  # Safely access possible missing keys
            rsp.results = results
            return rsp

    # TODO get rid of this once we completely switch to exec_async_v2
    def _exec_async_v1(
        self,
        database: str,
        engine: str,
        raw_code: str,
        inputs={},
        readonly=True,
    ):
        with debugging.span("transaction") as txn_span:
            if inputs:
                raise Exception(
                    "Inputs aren't currently supported using exec_async_v1 in SPCS."
                )

            with debugging.span("create_v1"):
                response = self._exec(
                    f"CALL {APP_NAME}.api.exec_async('{database}', '{engine}', ?, {readonly});",
                    raw_code,
                )
                if not response:
                    raise Exception("No results from exec_async")

            # grab the txn_id and state from the response
            first_row = next(iter(response))
            txn_id = first_row["ID"]
            state = first_row["STATE"]

            txn_span["txn_id"] = txn_id
            debugging.event("transaction_created", txn_span, txn_id=txn_id)
            # Wait for completion or failure
            if state != "COMPLETED":
                self._pending_transactions.append(txn_id)
                with debugging.span("wait", txn_id=txn_id):
                    poll_with_specified_overhead(
                        lambda: self._check_exec_async_status(txn_id), 0.2
                    )

            with debugging.span("fetch"):
                # List the result artifacts (and the URLs to retrieve them)
                artifact_urls = self._list_exec_async_artifacts(txn_id)

                return self._download_results(artifact_urls, txn_id, state)

    def _exec_async_v2(
        self,
        database: str,
        engine: str,
        raw_code: str,
        inputs={},
        readonly=True,
        nowait_durable=False,
        headers={},
    ):
        with debugging.span("transaction") as txn_span:
            with debugging.span("create_v2"):
                response = self._exec(
                    f"CALL {APP_NAME}.api.exec_async_v2('{database}','{engine}', ?, {inputs}, {readonly}, {nowait_durable}, {headers});",
                    raw_code,
                )

            if not response:
                raise Exception("Failed to create transaction")

            artifact_urls = {}
            rows = list(iter(response))

            # process the first row since txn_id and state are the same for all rows
            first_row = rows[0]
            txn_id = first_row['ID']
            state = first_row['STATE']
            filename = first_row['FILENAME']
            url = first_row['PRESIGNED_URL']

            txn_span["txn_id"] = txn_id
            debugging.event("transaction_created", txn_span, txn_id=txn_id)

            # fast path: transaction already finished
            if state in ["COMPLETED", "ABORTED"]:
                if txn_id in self._pending_transactions:
                    self._pending_transactions.remove(txn_id)
                artifact_urls[filename] = url

                # Process rows to get the rest of the artifacts
                for row in rows:
                    filename = row['FILENAME']
                    url = row['PRESIGNED_URL']
                    artifact_urls[filename] = url

            # Slow path: transaction not done yet; start polling
            else:
                self._pending_transactions.append(txn_id)
                with debugging.span("wait", txn_id=txn_id):
                    poll_with_specified_overhead(
                        lambda: self._check_exec_async_status(txn_id), 1.0
                    )
                artifact_urls = self._list_exec_async_artifacts(txn_id)

            with debugging.span("fetch"):
                return self._download_results(artifact_urls, txn_id, state)


    #--------------------------------------------------
    # Exec
    #--------------------------------------------------

    def exec_raw(
        self,
        database: str,
        engine: str,
        raw_code: str,
        inputs: dict = {},
        readonly=True,
        nowait_durable=False,
        headers={},
    ):
        raw_code = raw_code.replace("'", "\\'")
        use_exec_async_v2 = self.config.get("sf_use_exec_async_v2", True)

        if use_exec_async_v2:
            return self._exec_async_v2(
                database, engine, raw_code, inputs, readonly, nowait_durable, headers
            )
        else:
            return self._exec_async_v1(database, engine, raw_code, inputs, readonly)


    def format_results(self, results, task:m.Task|None=None) -> Tuple[DataFrame, List[Any]]:
        return result_helpers.format_results(results, task)

    #--------------------------------------------------
    # Exec format
    #--------------------------------------------------

    def exec_format(
        self,
        database: str,
        engine: str,
        raw_code: str,
        task: m.Task,
        format: str,
        inputs: dict = {},
        readonly=True,
        nowait_durable=False,
        headers={},
    ):
        # TODO: add headers and nowait_durable
        start = time.perf_counter()
        output_table = "out"+str(uuid.uuid4()).replace("-", "_")
        temp_table = f"temp_{output_table}"
        self._exec(f"call {APP_NAME}.api.exec_into(?, ?, ?, ?, ?);", [database, engine, raw_code, output_table, readonly])
        debugging.time("exec_format", time.perf_counter() - start, DataFrame())

        names = ", ".join([f"col{ix:03} as {name}" for (ix, name) in enumerate(task.return_cols())])
        self._exec(f"CREATE TEMPORARY TABLE {APP_NAME}.results.{temp_table} AS SELECT {names} FROM {APP_NAME}.results.{output_table};")
        self._exec(f"call {APP_NAME}.api.drop_result_table(?)", [output_table])

        temp = cast(snowflake.snowpark.DataFrame, self._exec(f"select * from {APP_NAME}.results.{temp_table}", raw=True))
        return (temp, [])

    #--------------------------------------------------
    # Custom model types
    #--------------------------------------------------

    def _get_ns(self, model:dsl.Graph):
        if model not in self._ns_cache:
            self._ns_cache[model] = Snowflake(model)
        return self._ns_cache[model]

    def to_model_type(self, model:dsl.Graph, name: str, source:str):
        parts = source.split(".")
        if len(parts) < 3:
            raise SnowflakeInvalidSource(Errors.call_source(), source)
        ns = self._get_ns(model)
        for part in parts:
            ns = getattr(ns, part)
        return ns

    #--------------------------------------------------
    # Transactions
    #--------------------------------------------------
    def txn_list_response_to_dicts(self, transactions):
        field_mapping = {
            "database_name": "database"
        }

        parsed_results = [
            {
                field_mapping.get(field, field): row[field.upper()]
                for field in LIST_TXN_SQL_FIELDS
            }
            for row in transactions
        ]
        return parsed_results

    def get_transaction(self, transaction_id):
        results = self._exec(
            f"CALL {APP_NAME}.api.get_transaction(?);", [transaction_id])
        if not results:
            return None

        txn = {field: results[0][field.upper()] for field in GET_TXN_SQL_FIELDS}

        created_on = txn.get("created_on")
        finished_at = txn.get("finished_at")
        if created_on:
            if finished_at:
                txn['duration'] = finished_at - created_on
            else:
                tz_info = created_on.tzinfo
                txn['duration'] = datetime.now(tz_info) - created_on
        return txn

    def list_transactions(self, **kwargs):
        id = kwargs.get("id", None)
        state = kwargs.get("state", None)
        limit = kwargs.get("limit", 100)
        only_active = kwargs.get("only_active", False)
        where_clause_arr = []

        if id:
            where_clause_arr.append(f"id = '{id}'")
        if state:
            where_clause_arr.append(f"state = '{state.upper()}'")
        else:
            if only_active:
                where_clause_arr.append("state in ('CREATED', 'RUNNING', 'PENDING')")

        if len(where_clause_arr):
            where_clause = f'WHERE {" AND ".join(where_clause_arr)}'
        else:
            where_clause = ""

        sql_fields = ", ".join([f.upper() for f in LIST_TXN_SQL_FIELDS])
        results = self._exec(f"select {sql_fields} from {APP_NAME}.api.transactions {where_clause} LIMIT ?", [limit])
        if not results:
            return []
        return self.txn_list_response_to_dicts(results)

    def cancel_transaction(self, transaction_id):
        self._exec(f"CALL {APP_NAME}.api.CANCEL_TRANSACTION(?);", [transaction_id])
        if transaction_id in self._pending_transactions:
            self._pending_transactions.remove(transaction_id)

    def cancel_pending_transactions(self):
        for txn_id in self._pending_transactions:
            self.cancel_transaction(txn_id)

    def get_transaction_events(self, txn_id: str, continuation_token:str=''):
        results = self._exec(
            f"SELECT {APP_NAME}.api.get_transaction_events(?, ?);",
            [txn_id, continuation_token],
        )
        if not results:
            return {
                "events": [],
                "continuation_token": None
            }
        row = results[0][0]
        return json.loads(row)

    #--------------------------------------------------
    # Snowflake specific
    #--------------------------------------------------

    def get_version(self):
        results = self._exec(f"SELECT {APP_NAME}.app.get_release()")
        if not results:
            return None
        return results[0][0]

    def list_warehouses(self):
        results = self._exec("SHOW WAREHOUSES")
        if not results:
            return []
        return [{"name":name}
                for (name, *rest) in results]

    def list_compute_pools(self):
        results = self._exec("SHOW COMPUTE POOLS")
        if not results:
            return []
        return [{"name":name, "status":status, "min_nodes":min_nodes, "max_nodes":max_nodes, "instance_family":instance_family}
                for (name, status, min_nodes, max_nodes, instance_family, *rest) in results]

    def list_valid_compute_pools_by_engine_size(self, engine_size:str):
        valid_pools = self.list_compute_pools()
        if not valid_pools:
            return []
        engine_pools = [item['name'] for item in valid_pools if item['status'] in VALID_POOL_STATUS and COMPUTE_POOL_MAP.get(item['instance_family']) == engine_size]
        if not engine_pools:
            return []
        return engine_pools

    def list_roles(self):
        results = self._exec("SELECT CURRENT_AVAILABLE_ROLES()")
        if not results:
            return []
        # the response is a single row with a single column containing
        # a stringified JSON array of role names:
        row = results[0]
        if not row:
            return []
        return [{"name": name} for name in json.loads(row[0])]

    def list_apps(self):
        all_apps = self._exec(f"SHOW APPLICATIONS LIKE '{RAI_APP_NAME}'")
        if not all_apps:
            all_apps = self._exec("SHOW APPLICATIONS")
            if not all_apps:
                return []
        return [{"name":name}
                for (time, name, *rest) in all_apps]

    def list_databases(self):
        results = self._exec("SHOW DATABASES")
        if not results:
            return []
        return [{"name":name}
                for (time, name, *rest) in results]

    def list_sf_schemas(self, database:str):
        results = self._exec(f"SHOW SCHEMAS IN {database}")
        if not results:
            return []
        return [{"name":name}
                for (time, name, *rest) in results]

    def list_tables(self, database:str, schema:str):
        results = self._exec(f"SHOW OBJECTS IN {database}.{schema}")
        items = []
        if results:
            for (time, name, db_name, schema_name, kind, *rest) in results:
                items.append({"name":name, "kind":kind.lower()})
        return items

    def schema_info(self, database:str, schema:str, tables:Iterable[str]):
        pks = self._exec(f"SHOW PRIMARY KEYS IN SCHEMA {database}.{schema};")
        fks = self._exec(f"SHOW IMPORTED KEYS IN SCHEMA {database}.{schema};")
        tables = ", ".join([f"'{x.upper()}'" for x in tables])
        columns = self._exec(textwrap.dedent(f"""
            select TABLE_NAME, COLUMN_NAME, DATA_TYPE
            from {database.upper()}.INFORMATION_SCHEMA.COLUMNS
            where TABLE_SCHEMA = '{schema.upper()}'
            and TABLE_NAME in ({tables})
            and TABLE_CATALOG = '{database.upper()}';
        """))
        results = defaultdict(lambda: {"pks": [], "fks": {}, "columns": {}})
        if pks:
            for row in pks:
                results[row[3].lower()]["pks"].append(row[4].lower()) # type: ignore
        if fks:
            for row in fks:
                results[row[7].lower()]["fks"][row[8].lower()] = row[3].lower()
        if columns:
            for row in columns:
                results[row[0].lower()]["columns"][row[1].lower()] = row[2].lower()
        return results

#--------------------------------------------------
# Snowflake Wrapper
#--------------------------------------------------

class PrimaryKey:
    pass

class Snowflake:
    def __init__(self, model, auto_import=False):
        self._model = model
        self._auto_import = auto_import
        if not isinstance(model._client.resources, Resources):
            raise ValueError("Snowflake model must be used with a snowflake config")
        self._dbs = {}
        imports = model._client.resources.list_imports(model=model.name)
        self._import_structure(imports)

    def _import_structure(self, imports: list[Import]):
        tree = self._dbs
        # pre-create existing imports
        schemas = set()
        for item in imports:
            database_name, schema_name, table_name = item["name"].lower().split('.')
            database = getattr(self, database_name)
            schema = getattr(database, schema_name)
            schemas.add(schema)
            schema._add(table_name, is_imported=True)
        for schema in schemas:
            schema._fetch_info()
        return tree

    def __getattribute__(self, __name: str) -> 'SnowflakeDB':
        if __name.startswith("_"):
            return super().__getattribute__(__name)
        __name = __name.lower()
        if __name in self._dbs:
            return self._dbs[__name]
        self._dbs[__name] = SnowflakeDB(self, __name)
        return self._dbs[__name]

class SnowflakeDB:
    def __init__(self, parent, name):
        self._name = name
        self._parent = parent
        self._model = parent._model
        self._schemas = {}

    def __getattribute__(self, __name: str) -> 'SnowflakeSchema':
        if __name.startswith("_"):
            return super().__getattribute__(__name)
        __name = __name.lower()
        if __name in self._schemas:
            return self._schemas[__name]
        self._schemas[__name] = SnowflakeSchema(self, __name)
        return self._schemas[__name]

class SnowflakeSchema:
    def __init__(self, parent, name):
        self._name = name
        self._parent = parent
        self._model = parent._model
        self._tables = {}
        self._imported = set()
        self._table_info = defaultdict(lambda: {"pks": [], "fks": {}, "columns": {}})

    def _fetch_info(self):
        self._table_info = self._model._client.resources.schema_info(self._parent._name, self._name, self._imported)

    def _add(self, name, is_imported=False):
        name = name.lower()
        if name in self._tables:
            return self._tables[name]
        if is_imported:
            self._imported.add(name)
        else:
            self._tables[name] = SnowflakeTable(self, name)
        return self._tables.get(name)

    def __getattribute__(self, __name: str) -> 'SnowflakeTable':
        if __name.startswith("_"):
            return super().__getattribute__(__name)
        table = self._add(__name)
        table._lazy_init()
        return table

class SnowflakeTable(dsl.Type):
    def __init__(self, parent, name):
        super().__init__(parent._model, f"sf_{name}")
        self._name = name
        self._model = parent._model
        self._parent = parent
        self._aliases = {}
        self._finalzed = False

    def _lazy_init(self):
        parent = self._parent
        name = self._name
        if name not in parent._imported:
            if self._parent._parent._parent._auto_import:
                with Spinner(f"Creating stream for {self.fqname()}", f"Stream created for {self.fqname()}"):
                    db_name = parent._parent._name
                    schema_name = parent._name
                    self._model._client.resources.create_import_stream(ImportSourceTable(db_name, schema_name, name), self._model.name)
                print("")
                parent._imported.add(name)
            else:
                imports = self._model._client.resources.list_imports(model=self._model.name)
                for item in imports:
                    cur_name = item["name"].lower().split(".")[-1]
                    parent._imported.add(cur_name)
            if name not in parent._imported:
                exception = SnowflakeImportMissingException(debugging.capture_code_info(), self.fqname(), self._model.name)
                raise exception from None

        if name not in parent._table_info:
            parent._fetch_info()

        self._finalize()

    def _finalize(self):
        if self._finalzed:
            return

        self._finalzed = True
        self._schema = self._parent._table_info[self._name]
        relation_name = self.fqname().replace(".", "_").lower()
        model:dsl.Graph = self._model
        model.install_raw(f"declare {relation_name}")

        with model.rule(dynamic=True):
            prop, id, val = std.Vars(3)
            if self._schema["pks"]:
                getattr(getattr(std.rel, relation_name), self._schema["pks"][0].upper())(id, val)
            else:
                getattr(std.rel, relation_name)(prop, id, val)
            self.add(snowflake_id=id)

        for prop, prop_type in self._schema["columns"].items():
            with model.rule(dynamic=True):
                id, val = std.Vars(2)
                getattr(getattr(std.rel, relation_name), prop.upper())(id, val)
                if getattr(self, prop.lower()).is_multi_valued:
                    inst = self(snowflake_id=id)
                    getattr(inst, prop.lower()).add(val)
                else:
                    self(snowflake_id=id).set(**{prop.lower(): val})

    def namespace(self):
        return f"{self._parent._parent._name}.{self._parent._name}"

    def fqname(self):
        return f"{self.namespace()}.{self._name}"

    def describe(self, **kwargs):
        model = self._model
        for k, v in kwargs.items():
            if v is PrimaryKey:
                self._schema["pks"] = [k]
            elif isinstance(v, tuple):
                (table, name) = v
                if isinstance(table, SnowflakeTable):
                    fk_table = table
                    pk = fk_table._schema["pks"]
                    with model.rule():
                        inst = fk_table()
                        me = self()
                        getattr(inst, pk[0]) == getattr(me, k)
                        if getattr(self, name).is_multi_valued:
                            getattr(me, name).add(inst)
                        else:
                            me.set(**{name: inst})
                else:
                    raise ValueError(f"Invalid foreign key {v}")
            else:
                raise ValueError(f"Invalid column {k}={v}")
        return self

#--------------------------------------------------
# Graph
#--------------------------------------------------

def Graph(
    name,
    *,
    profile: str | None = None,
    config: Config,
    dry_run: bool = False,
    isolated: bool = True,
    connection: Session | None = None,
    keep_model: bool = False,
    nowait_durable: bool = True,
    format: str = "default",
):
    client = Client(
        Resources(profile, config, connection),
        rel.Compiler(config),
        name,
        config,
        dry_run=dry_run,
        isolated=isolated,
        keep_model=keep_model,
        nowait_durable=nowait_durable
    )
    client.install("pyrel_base", dsl.build.raw_task("""
        @inline
        def make_identity(x..., z):
            exists((u) | hash128({x...}, x..., u) and
            hash_value_uint128_convert(u, z))

        @inline
        def pyrel_default({F}, c, k..., v):
            F(k..., v) or (not F(k..., _) and v = c)

        @inline
        def pyrel_unwrap(x in UInt128, y): y = x

        @inline
        def pyrel_dates_period_days(x in Date, y in Date, z):
            exists((u) | dates_period_days(x, y , u) and u = ::std::common::^Day[z])

        @inline
        def pyrel_datetimes_period_milliseconds(x in DateTime, y in DateTime, z in Int):
            exists((u) | datetimes_period_milliseconds(x, y , u) and u = ^Millisecond[z])
    """))
    return dsl.Graph(client, name, format=format)



from pathlib import Path
import random
import os

import pytest
import snowflake.connector

import relationalai as rai
from relationalai import debugging
from relationalai.clients.profile_polling import TransactionEventsFeedbackHandler
from relationalai.debugging import logger
from relationalai.clients import config as cfg
from relationalai.util.snowflake_logger import SnowflakeLogger
from relationalai.util.span_format_test import SpanCollectorHandler, assert_valid_span_structure
from testutil.snapshot import path_to_slug, validate_block_snapshots

def engine_config_fixture(size):
    # Check for an externally provided engine name
    # It is used in GitHub Actions to run tests against a specific engine
    engine_name = os.getenv("ENGINE_NAME")
    if engine_name:
        # If engine name was provided, just yield the config
        config = make_config(engine_name)
        yield config
        return

    # If there's a local config file, use it, including
    # the engine specified there.
    config = cfg.Config()
    if config.file_path is not None:
        yield config
        return

    # Otherwise, create a new engine and delete it afterwards.
    random_number = random.randint(1000000000, 9999999999)
    engine_name = f"pyrel_test_{random_number}"
    create_engine(engine_name, size=size)

    yield make_config(engine_name)

    delete_engine(engine_name)

def create_engine(engine_name: str, size: str):
    config = make_config(engine_name)

    sf_compute_pool = os.getenv("SF_TEST_COMPUTE_POOL", config.get("compute_pool", ""))

    provider = rai.Resources(config=config)
    print(f"Creating engine {engine_name}")
    provider.create_engine(name=engine_name, size=size, pool=sf_compute_pool)
    print(f"Engine {engine_name} created")


def delete_engine(engine_name: str):
    print(f"Deleting engine {engine_name}")
    config = make_config(engine_name)
    provider = rai.Resources(config=config)
    provider.delete_engine(engine_name)
    print(f"Engine {engine_name} deleted")

def make_config(engine_name: str) -> cfg.Config:
    cloud_provider = os.getenv("RAI_CLOUD_PROVIDER")

    match cloud_provider:
        case None:
            raise ValueError("RAI_CLOUD_PROVIDER must be set")
        case "azure":
            client_id = os.getenv("RAI_CLIENT_ID")
            client_secret = os.getenv("RAI_CLIENT_SECRET")
            if client_id is None or client_secret is None:
                raise ValueError(
                    "RAI_CLIENT_ID, RAI_CLIENT_SECRET must be set if RAI_CLOUD_PROVIDER is set to 'azure'"
                )
            # Running against prod
            return cfg.Config(
                {
                    "platform": "azure",
                    "host": "azure.relationalai.com",
                    "port": "443",
                    "region": "us-east",
                    "scheme": "https",
                    "client_credentials_url": "https://login.relationalai.com/oauth/token",
                    "client_id": client_id,
                    "client_secret": client_secret,
                    "engine": engine_name,
                }
            )
        case "snowflake":
            sf_username = os.getenv("SF_TEST_ACCOUNT_USERNAME")
            sf_password = os.getenv("SF_TEST_ACCOUNT_PASSWORD")
            sf_account = os.getenv("SF_TEST_ACCOUNT_NAME")
            sf_warehouse = os.getenv("SF_TEST_WAREHOUSE_NAME")
            sf_app_name = os.getenv("SF_TEST_APP_NAME")
            sf_compute_pool = os.getenv("SF_TEST_COMPUTE_POOL")
            sf_use_exec_async_v2 = os.getenv("SF_USE_EXEC_ASYNC_V2", False)
            if sf_username is None or sf_password is None:
                raise ValueError(
                    "SF_TEST_ACCOUNT_USERNAME, SF_TEST_ACCOUNT_PASSWORD, SF_TEST_ACCOUNT_NAME must be set if RAI_CLOUD_PROVIDER is set to 'snowflake'"
                )
            return cfg.Config(
                {
                    "platform": "snowflake",
                    "user": sf_username,
                    "password": sf_password,
                    "account": sf_account,
                    "role": "rai_integration_consumer",
                    "warehouse": sf_warehouse,
                    "rai_app_name": sf_app_name,
                    "engine": engine_name,
                    "compute_pool": sf_compute_pool,
                    "sf_use_exec_async_v2": sf_use_exec_async_v2
                }
            )
        case _:
            raise ValueError(f"Unsupported cloud provider: {cloud_provider}")

def snowflake_logger_fixture():
    if not os.getenv("SF_REPORTING_PASSWORD"):
        print('benchmarking logger disabled since config env vars not present')
        yield
        return

    conn = snowflake.connector.connect(
        user=os.getenv("SF_REPORTING_USER"),
        password=os.getenv("SF_REPORTING_PASSWORD"),
        account=os.getenv("SF_REPORTING_ACCOUNT"),
        role=os.getenv("SF_REPORTING_ROLE"),
        warehouse=os.getenv("SF_REPORTING_WAREHOUSE"),
        database=os.getenv("SF_REPORTING_DATABASE"),
        schema=os.getenv("SF_REPORTING_SCHEMA"),
    )
    snowflake_logger = SnowflakeLogger(conn)
    logger.addHandler(snowflake_logger)
    yield
    snowflake_logger.shut_down()

def profile_poller_fixture(engine_config):
    resources = rai.Resources(config=engine_config)
    handler = TransactionEventsFeedbackHandler(resources)
    logger.addHandler(handler)
    yield

def span_structure_validator_fixture():
    handler = SpanCollectorHandler()
    logger.addHandler(handler)
    yield handler
    logger.removeHandler(handler)
    assert_valid_span_structure(handler.nodes, handler.events)

def root_span_fixture():
    git_info = get_git_info()
    attrs = {
        **git_info,
        "cloud_provider": os.getenv("RAI_CLOUD_PROVIDER"),
    }
    with debugging.span("test_session", **attrs):
        yield

def get_git_info():
    if os.getenv("GITHUB_REF_NAME") is not None:
        return {
            'commit': os.getenv("GITHUB_SHA"),
            'branch': os.getenv("GITHUB_REF_NAME"),
        }
    else:
        # shell out to get the info
        return {
            'commit': 'local',
            'branch': 'local',
        }

def randomize(name: str) -> str:
    return f"{name}_{random.randint(1000000000, 9999999999)}"

def do_snapshot_test(
        config: cfg.Config,
        engine_size: str,
        snapshot,
        file_path: Path,
        runs: int = 1,
        span_name: str = "test",
        create_db: bool = True,
    ):
    """
    Run the PyRel program in the given `file_path`, testing the results against a snapshot file,
    using the given `config` and `engine_size`.
    """

    test_name = file_path.stem

    provider = rai.Resources(config=config)

    def get_results(event):
        if "results" in event:
            return str(event["results"]["values"])

    with debugging.span(span_name, name=test_name, engine_size=engine_size):
        for i in range(runs):
            with debugging.span("run", idx=i):
                with debugging.span("program", main=os.path.relpath(file_path), platform=config.get("platform", None)) as span:
                    db_name = randomize(test_name)
                    validate_block_snapshots(file_path, snapshot, get_results, "query", {
                        "name": db_name,
                        "config": config,
                        "span": span,
                    })
                    # Some tests always use the same DB, never creating a DB named `db_name`,
                    # so we don't need to delete it.
                    if create_db:
                        try:
                            provider.delete_graph(db_name)
                        except Exception as e:
                            print(f"Failed to delete graph {db_name}: {e}")

def all_files_in_dir(test_case_dir: Path):
    test_case_files = [path for path in test_case_dir.rglob("*.py")]
    return pytest.mark.parametrize(
        "file_path", test_case_files, ids=lambda path: path_to_slug(path, test_case_dir)
    )



#pyright: reportUnusedExpression=false
from typing import Tuple
import relationalai as rai

# Since exports aren't supported in Azure, we'll do this as a dry run
# to test just the bytecode generation
model = rai.Model(name=globals().get("name", ""), config=globals().get("config"), dry_run=True)

Person = model.Type("Person")
Adult = model.Type("Adult")

with model.rule():
    Person.add(name="Joe", age=74)
    Person.add(name="Bob", age=40)
    Person.add(name="Jane", age=10)

with model.rule():
    p = Person()
    p.age >= 18
    p.set(Adult)

@model.export("sandbox.public")
def test_export() -> Tuple[str, int]:
    a = Adult()
    return a.name, a.age



from relationalai import metamodel as mm
from gentest.gen import ir
from gentest.gen.action import gen_task_body

from gentest.gen.scope import GenScope

def mk_call(op: mm.Task, args: list[mm.Var]):
    return mm.Action(mm.ActionType.Call, mm.Var(mm.Builtins.Task, value=op), [], {prop: arg for prop, arg in zip(op.properties, args)})


def hydrate_call(scope: GenScope, call: ir.Call, into: list[mm.Action]) -> mm.Var|None:
    args = [hydrate_expr(scope, arg, into) for arg in call.args]
    into.append(mk_call(call.op, args))
    if not call.op.properties[-1].is_input:
        return args[-1]

def hydrate_expr(scope: GenScope, expr: ir.Expr|list[ir.Expr], into: list[mm.Action]) -> mm.Var:
    match expr:
        case mm.Var():
            return expr
        case ir.Constant():
            return mm.Var(expr.type, value=expr.value)
        case ir.Call():
            ret = hydrate_call(scope, expr, into)
            assert ret
            return ret
        case list():
            # @FIXME: Correct typing for list vars?
            return mm.Var(mm.Builtins.Any, value=[hydrate_expr(scope, v, into) for v in expr])
        case _:
            raise Exception(f"Unhandled expression type {expr.__class__.__name__}")

def as_mm_action_type(op: ir.AssertOp) -> mm.ActionType:
    match op:
        case ir.AssertOp.Bind:
            return mm.ActionType.Bind
        case ir.AssertOp.Persist:
            return mm.ActionType.Persist
        case ir.AssertOp.Unpersist:
            return mm.ActionType.Unpersist
        case _:
            raise Exception(f"Unhandled action type {op}")

def hydrate_action(scope: GenScope, action: ir.Action, into: list[mm.Action]):
    match action:
        case ir.EntityProvider():
            into.append(mm.Action(mm.ActionType.Get, action.entity, [action.entity.type], action.properties))
        case ir.ComputedProvider():
            hydrate_call(scope, action.call, into)
        case ir.AssertType():
            into.append(mm.Action(as_mm_action_type(action.op), action.entity, [action.type]))
        case ir.AssertProperty():
            into.append(mm.Action(as_mm_action_type(action.op), action.entity, [], {action.property: hydrate_expr(scope, action.value, into)}))
        case ir.Call():
            hydrate_call(scope, action, into)
        case _:
            raise Exception(f"Unhandled action type {action.__class__.__name__}.")

def hydrate_actions(scope: GenScope, actions: list[ir.Action]):
    hydrated_actions = []
    for action in actions:
        hydrate_action(scope, action, hydrated_actions)

    return hydrated_actions, scope

def hydrate(pair: tuple[list[ir.Action], GenScope]):
    hydrated_actions, _ = hydrate_actions(pair[1], pair[0])
    return mm.Task(behavior=mm.Behavior.Query, items=hydrated_actions)

def gen_task(root_scope: GenScope):
    return gen_task_body(root_scope, 5).map(hydrate)




import relationalai as rai
from relationalai.std.graphs import Graph

model = rai.Model(name=globals().get("name", "test_weighted"), config=globals().get("config"))


def run_test(algo, ix: int, raw_edges):
    Node = model.Type(f"Node{ix}")
    graph = Graph(model, weighted=ix > 0)

    # Install graph data
    with model.rule(dynamic=True):
        for _from, _to, *weights in raw_edges:
            weight = weights[ix - 1] if ix > 0 else None
            Node.add(id=_from)
            Node.add(id=_to)
            if weight is None:
                graph.Edge.add(_from, _to)
            else:
                graph.Edge.add(_from, _to, weight=weight)

    # Test algo
    with model.query() as select:
        n = Node()
        res = select(n.id, getattr(graph.compute, algo)(n.id))

    print(res.results)

def run_tests(algo, raw_edges):
    case_count = len(raw_edges[0]) - 1
    for ix in range(0, case_count):
        run_test(algo, ix, raw_edges)

# Format (from, to, weight1, weight2, ...)

# Louvain
run_tests("louvain", [
    # First embedded 3-clique.
    (1, 2, 1.0, 0.0),
    (1, 3, 1.0, 1.0),
    (2, 3, 1.0, 0.0),

    # Second embedded 3-clique.
    (1, 4, 1.0, 1.0), (4, 5, 1.0, 10.0), (4, 6, 1.0, 1.0),

    # Connection between the embedded 3-cliques.
    (1, 4, 1.0, 1.0),

    # Weaker edges connecting the 6-clique in full.
    (1, 5, 0.2, 0.2),
    (1, 6, 0.2, 9.0),


    (2, 4, 0.2, 0.2),
    (2, 5, 0.2, 0.2),
    (2, 6, 0.2, 0.2),

    (3, 4, 0.2, 0.2),
    (3, 5, 0.2, 0.2),
    (3, 6, 0.2, 0.2)
])

# Degree
# Ensure unweighted degree still works properly on weighted graphs
run_tests("degree", [
    (11, 12, 1.0, 1.0),
    (12, 13, 1.0, -1.0),
    (13, 13, 1.0, 0.0),
    (12, 14, 1.0, 9.0)
])



from collections import OrderedDict
import io
import re
import textwrap
import time
from typing import List, Dict, Tuple, Any, Iterable, cast
from .metamodel import Agent, Base, Namer, Task, Var, Value, Property, Action, ActionType, AllItems, Behavior, Builtins, Type
from . import debugging
from .clients import test
from .clients.config import Config
import random
import rich
import json

DEBUG = False

#--------------------------------------------------
# SortedSet
#--------------------------------------------------

class SortedSet:
    def __init__(self, elements=None):
        # Using OrderedDict to maintain insertion order and uniqueness
        self._elements = OrderedDict((element, None) for element in elements) if elements is not None else OrderedDict()

    def add(self, element):
        # Add element only if it does not exist, maintaining insertion order
        self._elements[element] = None

    def remove(self, element):
        # Remove element if it exists
        if element in self._elements:
            del self._elements[element]

    def __contains__(self, element):
        # Check if element is in the set
        return element in self._elements

    def __iter__(self):
        # Allow iteration over elements in their insertion order
        return iter(self._elements)

    def __len__(self):
        # Return the number of unique elements
        return len(self._elements)

    def __str__(self):
        # String representation of the sorted set
        return "SortedSet(" + str(list(self._elements.keys())) + ")"

#--------------------------------------------------
# PyRel emitter
#--------------------------------------------------

op_mapping = {
    "=": "==",
    "^": "**"
}

class PyRelEmitter:
    def __init__(self):
        self.namer = Namer()
        self.bound = {}
        self.used = set()
        self.used_types = set()

    def reset(self, full=False):
        self.namer.reset()
        self.bound.clear()
        self.used.clear()
        if full:
            self.used_types.clear()

    #--------------------------------------------------
    # Helpers
    #--------------------------------------------------

    def sanitize(self, input_string):
        # Replace non-alphanumeric characters with underscores
        sanitized = re.sub(r'[^:\w]|^(?=\d)', '_', input_string)

        # Check if the resulting string is a keyword and append an underscore if it is
        if sanitized in ["from", "is", "not", "or", "and", "as", "assert", "break", "class", "continue"]:
            sanitized += '_'

        return sanitized

    #--------------------------------------------------
    # Emit
    #--------------------------------------------------

    def emit(self, item:AllItems) -> str:
        match item:
            case Task():
                return self.emit_task(item)
            case Type():
                return self.emit_type(item)
            case Var():
                return self.emit_var(item)
            case Action():
                return self.emit_action(item)
            case Property():
                return self.emit_property(item)
        raise Exception(f"Unknown item type: {type(item)}")

    #--------------------------------------------------
    # Val handling
    #--------------------------------------------------

    def emit_val(self, value: Value) -> str:
        if value is True or value is False:
            return str(value)
        if isinstance(value, list):
            return f"{', '.join([self.emit_var(v) for v in value])}"
        if isinstance(value, bytes):
            return f"0x{value.hex()}"
        if isinstance(value, Property):
            return self.sanitize(value.name)
        if isinstance(value, Type):
            return self.sanitize(value.name)
        return json.dumps(value)

    #--------------------------------------------------
    # Var handling
    #--------------------------------------------------

    def emit_var_name(self, v:Var) -> str:
        name = self.namer.get(v)
        return name.lower()

    def emit_var(self, var: Var|Value) -> str:
        if not isinstance(var, Var):
            return self.emit_val(var)
        if var.value is not None:
            return self.emit_val(var.value)
        if var.name == "_":
            return "_"
        return self.emit_var_name(var)

    #--------------------------------------------------
    # Types and Properties
    #--------------------------------------------------

    def emit_type(self, t:Type) -> str:
        self.used_types.add(t)
        return self.sanitize(t.name)

    def emit_property(self, p:Property) -> str:
        return self.sanitize(p.name)

    #--------------------------------------------------
    # Actions
    #--------------------------------------------------

    def action_params(self, a:Action, extras:List[str]) -> str:
        params = []
        types = a.types
        if len(types):
            if a.action == ActionType.Get:
                types = types[1:]
            if len(types):
                params.append(f"{', '.join([self.emit(t) for t in types])}")
        for k,v in a.bindings.items():
            if v.value is None and v not in self.bound:
                self.bound[v] = self.emit_var(v)
                extras.append(f"{self.emit_var(v)} = {self.emit_var(a.entity)}.{self.emit(k)}")
                continue
            if v in self.bound:
                v = self.bound[v]
            else:
                v = self.emit(v)
            params.append(f"{self.emit(k)}={v}")
        return ", ".join(params)

    def emit_action(self, a:Action) -> str:
        if a.action == ActionType.Call:
            return self.emit_call(a)
        else:
            var = self.emit_var_name(a.entity)
            if a.entity not in self.bound:
                self.bound[a.entity] = var
            root = self.emit(a.types[0]) if len(a.types) else var
            extras = []
            params = self.action_params(a, extras)
            line = ""
            if a.action == ActionType.Get:
                line = f"{var} = {root}({params})"
            elif a.action == ActionType.Bind:
                line = f"{var}.set({params})"
            elif a.action == ActionType.Persist:
                line = f"{var}.persist({params})"
            elif a.action == ActionType.Unpersist:
                line = f"{var}.unpersist({params})"
            if not len(extras):
                return line
            extras_str = '\n'.join(extras)
            return f"{line}\n{extras_str}"

    #--------------------------------------------------
    # Action: Call
    #--------------------------------------------------

    def emit_call(self, a:Action) -> str:
        if a.entity.isa(Builtins.Infix):
            name = cast(Type, a.entity.value).name
            name = op_mapping.get(name, name)
            args = [self.emit(v) for v in a.bindings.values()]
            if len(args) == 2:
                return f"{args[0]} {name} {args[1]}"
            else:
                return f"{args[2]} = {args[0]} {name} {args[1]}"
        raise Exception(f"Unsupported call type: {a.entity}")

    #--------------------------------------------------
    # Tasks
    #--------------------------------------------------

    def emit_task_items(self, task:Task) -> str:
        items = [self.emit(i) for i in task.items]
        if not len(items):
            items.append("pass")
        return textwrap.indent("\n".join(items), "    ")

    def emit_task(self, task:Task) -> str:
        if task.behavior == Behavior.Query:
            head = "with model.query() as select:"
            return f"{head}\n{self.emit_task_items(task)}"
        return "TASK"

    #--------------------------------------------------
    # Documents
    #--------------------------------------------------

    def emit_header(self) -> str:
        types = []
        for t in self.used_types:
            t_name = self.emit(t)
            types.append(f"{t_name} = model.Type(\"{t_name}\")")
        lines = [
            "import relationalai as rai",
            "",
            "model = rai.clients.test.Graph(\"gentest\", config=config)",
            "",
            *types
        ]
        return "\n".join(lines)

    def emit_doc(self, tasks:List[Task]) -> str:
        task_code = []
        for t in tasks:
            self.reset()
            task_code.append(self.emit(t))
        body = "\n\n".join(task_code)
        return f"{self.emit_header()}\n\n{body}"

#--------------------------------------------------
# GenStats
#--------------------------------------------------

class GenStats:
    def __init__(self, debug=DEBUG):
        self.start = time.perf_counter()
        self.runs = 0
        self.failures = []
        self.debugs = []
        self.gen_time = 0.0
        self.emit_time = 0.0
        self.exec_time = 0.0
        self._debugging = debug

    def success(self):
        self.runs += 1

    def fail(self, seed, expected, actual, mismatches, code, exception=None):
        self.runs += 1
        self.failures.append((seed, expected, actual, mismatches, code, exception))

    def debug(self, seed, expected, actual, code):
        if self._debugging:
            self.debugs.append((seed, expected, actual, code))

    def finish(self):
        self.elapsed = time.perf_counter() - self.start
        return self

    def __rich__(self) -> str:
        failures = len(self.failures)
        results_str = f"[red bold]{failures:,.0f}[/red bold]  [green bold]{self.runs-failures:,.0f}[/green bold]"
        def millis(s, color="white"):
            return f"[{color} bold]{s * 1000:,.0f}[/{color} bold][{color}]ms[/{color}]"
        summary = f"{results_str}  {millis(self.elapsed, 'yellow')}  [dim]|  {millis(self.gen_time)} gen  {millis(self.emit_time)} emit  {millis(self.exec_time)} exec[/dim]"
        return summary

    def print_failures(self, count=None):
        if count is None:
            count = len(self.failures)
        for seed, expected, actual, mismatches, code, exception in self.failures[0:count]:
            rich.print("[red bold]---------------------------------------------------------------")
            rich.print(f"[red bold]SEED: {seed}")
            if exception:
                print("")
                print(exception)
                rich.print("[dim]---------------------------------------------------------------")
            if len(mismatches):
                for mismatch in mismatches:
                    rich.print(f"[red bold]{mismatch[0]}")
                    rich.print(f"[red bold]{mismatch[1]}")
                rich.print("")
                rich.print("[dim]---------------------------------------------------------------")
            print("")
            print(expected)
            print("")
            rich.print("[dim]---------------------------------------------------------------")
            print("")
            print(actual)
            print("")
            rich.print("[dim]---------------------------------------------------------------")
            print("")
            rich.print(code)
            print("")

    def print_debugs(self):
        for seed, expected, actual, code in self.debugs:
            rich.print("[green bold]---------------------------------------------------------------")
            rich.print(f"[green bold]SEED: {seed}")
            print("")
            print(expected)
            rich.print("[dim]---------------------------")
            print("")
            print(code)
            print("")

#--------------------------------------------------
# GenProfile
#--------------------------------------------------

class GenProfile:
    def __init__(self):

        self.identifier_len = (3, 6)
        self.string_len = (3, 5)
        self.number_range = (0, 1000000)

        self.types = (1, 20)
        self.type_properties = (0, 10)

        self.task_actions = (1, 10)

        self.get_weight = 1
        self.get_types = (0, 4)
        self.get_properties = (0, 4)

        self.bind_weight = 1
        self.bind_types = (0, 4)
        self.bind_properties = (0, 4)

        self.call_weight = 1
        self.aggregate_weight = 1

        self.quantifier_weight = 0.2
        self.sub_task_weight = 0.2

#--------------------------------------------------
# Gen
#--------------------------------------------------

class Gen:
    def __init__(self, seed, profile:GenProfile = GenProfile()):
        self.seed = seed
        self.random = random.Random(seed)
        self.eq = Eq()
        self.emitter = PyRelEmitter()
        self.profile = profile
        self.config = Config()

        self.types = []
        for _ in range(self.rand_range(self.profile.types)):
            self.types.append(self.type())
        for t in self.types:
            t.properties = [self.property() for _ in range(self.rand_range(self.profile.type_properties))]

    #--------------------------------------------------
    # Helpers
    #--------------------------------------------------

    def rand_range(self, r):
        return self.random.randint(r[0], r[1])

    def choice(self, enum):
        return self.random.choice(list(enum))

    def choices(self, enum, k):
        options = list(enum)
        return self.random.sample(enum, k=min(k, len(options)))

    #--------------------------------------------------
    # Primitives
    #--------------------------------------------------

    def string(self, prefix="", length=None):
        length = length or self.rand_range(self.profile.string_len)
        return self.emitter.sanitize(prefix + ''.join(self.random.choice('abcdefghijklmnopqrstuvwxyz') for _ in range(length)))

    def identifier(self, length=None):
        length = length or self.rand_range(self.profile.identifier_len)
        return self.string("", length)

    def number(self):
        return self.random.randint(0, 1000000)

    def value(self, allowed_types=None) -> Value:
        if not allowed_types:
            allowed_types = [self.number, self.string]
        v = self.random.choice(allowed_types)
        return cast(Value, v())

    #--------------------------------------------------
    # Types and Properties
    #--------------------------------------------------

    def type(self) -> Type:
        name = self.identifier().capitalize()
        properties = []  # [self.generate_random_property() for _ in range(self.random.randint(1, 3))]
        parents = []  # Could recursively add random Types if necessary, but keeping it simple here
        return Type(name=name, properties=properties, parents=parents)

    def property(self) -> Property:
        name = self.identifier()
        type = self.choice(self.types)
        is_input = False
        return Property(name=name, type=type, is_input=is_input)

    def _choose_prop(self, types:List[Type]) -> Property|None:
        type = self.random.choice(types)
        if len(type.properties):
            return self.random.choice(type.properties)

    def prop_value(self, vars:SortedSet, allow_self=True, allowed_types=None) -> Var:
        i = self.random.randint(1, 10)
        if i < 5 and len(vars):
            return self.choice(vars)
        if i < 8 and allow_self:
            return self.var()
        return Var(value=self.value(allowed_types=allowed_types))

    #--------------------------------------------------
    # Var
    #--------------------------------------------------

    def var(self) -> Var:
        # type = self.choice(self.types)
        type = Builtins.Any
        name = None # self.string("Var_") if self.random.choice([True, False]) else None
        value = None  # Could implement logic to generate random Value based on Type
        return Var(type=type, name=name, value=value)

    #--------------------------------------------------
    # Actions: Get
    #--------------------------------------------------

    def action_get(self, vars:SortedSet, sets: Dict[Tuple[Var, Property], Var]) -> Action:
        entity = self.var()
        types = self.choices(self.types, self.rand_range(self.profile.get_types))
        if len(types):
            entity.type = types[0]
        props = SortedSet()
        all_opts = [entity.type, *types]
        for _ in range(self.rand_range(self.profile.get_properties)):
            prop = self._choose_prop(all_opts)
            if prop:
                props.add(prop)
        bindings = {prop: self.prop_value(vars) for prop in props}
        # We can ref our own vars, so we add them to the pool after the fact
        if len(props) or len(types):
            vars.add(entity)
        for k, v in bindings.items():
            if v.value is None:
                vars.add(v)
        return Action(action=ActionType.Get, entity=entity, types=types, bindings=bindings)

    #--------------------------------------------------
    # Actions: Bind
    #--------------------------------------------------

    def bind_var(self, vars:SortedSet) -> Var:
        available = [v for v in vars if v.type in self.types]
        if len(available):
            return self.choice(available)
        return self.var()

    def action_bind(self, vars:SortedSet, sets: Dict[Tuple[Var, Property], Var]) -> Action:
        entity = self.bind_var(vars)
        types = self.choices(self.types, self.rand_range(self.profile.bind_types))
        if len(types):
            entity.type = types[0]
        props = SortedSet()
        all_opts = [entity.type, *types]
        for _ in range(self.rand_range(self.profile.bind_properties)):
            prop = self._choose_prop(all_opts)
            if prop:
                props.add(prop)
        bindings = {prop: self.prop_value(vars, False) for prop in props}
        for k, v in bindings.items():
            sets[(entity, k)] = v

        # We can ref our own vars, so we add them to the pool after the fact
        return Action(action=ActionType.Bind, entity=entity, types=types, bindings=bindings)

    #--------------------------------------------------
    # Actions: Call
    #--------------------------------------------------

    def action_call(self, vars:SortedSet, sets: Dict[Tuple[Var, Property], Var]) -> Action:
        assert len(vars)

        op:Type = self.choice([
            Builtins.gt,
            Builtins.gte,
            Builtins.lt,
            Builtins.lte,
            Builtins.eq,
            Builtins.neq,
            Builtins.plus,
            Builtins.minus,
            Builtins.mult,
            Builtins.div,
            Builtins.pow
        ])
        bindings = {}
        last_input = None
        has_non_value = False
        for prop in op.properties:
            if prop.is_input:
                bindings[prop] = self.prop_value(vars, False, allowed_types=[self.number])
                if bindings[prop].value is None:
                    has_non_value = True
                last_input = prop
            else:
                bindings[prop] = self.var()
        if not has_non_value and last_input:
            bindings[last_input] = [v for v in vars if v.value is None][0]

        return Action(action=ActionType.Call, entity=Var(value=op), bindings=bindings)

    #--------------------------------------------------
    # Task
    #--------------------------------------------------

    def task(self) -> Task:
        behavior = self.choice([Behavior.Query])
        items = []
        vars = SortedSet()
        sets = {}
        for ix in range(self.rand_range(self.profile.task_actions)):
            if ix <= 1 or not len(vars):
                items.append(self.action_get(vars, sets))
            else:
                items.append(self.random.choice([
                    self.action_get,
                    self.action_bind,
                    self.action_call
                ])(vars, sets))
        return Task(behavior=behavior, items=items)

    #--------------------------------------------------
    # Tests
    #--------------------------------------------------

    def test_task(self, seed, stats:GenStats):
        gen_start = time.perf_counter()
        self.random.seed(seed)
        self.emitter.reset(True)
        self.eq.reset()

        expected = self.task()
        expected.normalize()
        stats.gen_time += time.perf_counter() - gen_start

        if not len(expected.items):
            stats.success()
            return

        emit_start = time.perf_counter()
        code = self.emitter.emit_doc([expected])
        stats.emit_time += time.perf_counter() - emit_start
        actual = None
        exception = None

        try:
            exec_time = time.perf_counter()
            exec(code, {"config": self.config}, None)
            result = test.proxied_clients.pop()
            if len(result.blocks):
                # When use_value_types is active, the first task is installing all the types
                if self.config.get("compiler.use_value_types", None):
                    actual = result.blocks[1].task
                else:
                    actual = result.blocks[0].task
                actual.normalize()
            stats.exec_time += time.perf_counter() - exec_time

        except Exception:
            output = io.StringIO()
            console = rich.console.Console(file=output, force_terminal=True)
            console.print_exception(extra_lines=0)
            exception = output.getvalue()

        stats.debug(seed, expected, actual, code)

        if not self.eq.eq(expected, actual):
            stats.fail(seed, expected, actual, self.eq.mismatches[:], code, exception)
        else:
            stats.success()

#--------------------------------------------------
# Eq
#--------------------------------------------------

OP_INVERSES = { ">": "<", ">=": "<=", "<": ">", "<=": ">=" }
OP_COMMUTATIVE = { "+", "*", "=", "!=" }

class Eq():
    def __init__(self):
        self.mapping = {}
        self.mismatches = []

    def reset(self):
        self.mapping.clear()
        self.mismatches.clear()

    #--------------------------------------------------
    # eq
    #--------------------------------------------------

    def eq(self, a:AllItems|Base|Value|None, b:AllItems|Base|Value|None):
        if type(a) != type(b):
            return False
        if isinstance(a, Task) and isinstance(b, Task):
            return self.task(a, b)
        if isinstance(a, Type) and isinstance(b, Type):
            return self.type(a, b)
        if isinstance(a, Property) and isinstance(b, Property):
            return self.property(a, b)
        if isinstance(a, Agent) and isinstance(b, Agent):
            return self.agent(a, b)
        if isinstance(a, Var) and isinstance(b, Var):
            return self.var(a, b)
        if isinstance(a, Action) and isinstance(b, Action):
            return self.action(a, b)
        if isinstance(a, dict) and isinstance(b, dict):
            return self.eq_dict(a, b)
        if isinstance(a, list) and isinstance(b, list):
            return self.eq_list(a, b)

        return a == b

    #--------------------------------------------------
    # Collection eq
    #--------------------------------------------------

    def eq_dict(self, a:Dict[str,Any], b:Dict[str,Any]):
        b_str_ks = {str(k): v for k,v in b.items()}
        a_str_ks = {str(k): v for k,v in a.items()}
        for k,v in a.items():
            if not self.eq(v, b_str_ks.get(str(k))):
                return False
        for k in b_str_ks.keys():
            if k not in a_str_ks:
                return False
        return True

    def eq_list(self, a:List[Any], b:List[Any]):
        for i in range(len(a)):
            if not self.eq(a[i], b[i]):
                return False
        for extra in b[len(a):]:
            return False
        return True

    def eq_attrs(self, a:AllItems, b:AllItems, attrs:List[str]):
        for attr in attrs:
            if not self.eq(getattr(a, attr), getattr(b, attr)):
                return False
        return True

    def eq_set(self, a:Iterable[Any], b:Iterable[Any]):
        for item in a:
            found = False
            for b_item in b:
                if self.eq(item, b_item):
                    found = True
                    break
            if not found:
                return False
        return True

    #--------------------------------------------------
    # Metamodel eq
    #--------------------------------------------------

    def type(self, a:Type, b:Type):
        return self.eq_attrs(a, b, ['name'])

    def property(self, a:Property, b:Property):
        # return self.eq_attrs(a, b, ['name', 'type', 'is_input'])
        return self.eq_attrs(a, b, ['name'])

    def task(self, task:Task, b:Task):
        return self.eq_attrs(task, b, ["name", "parents", "behavior", "items"])
        # return self.eq_attrs(task, b, ["name", "bindings", "properties", "parents", "inline", "behavior", "items"])

    def agent(self, a:Agent, b:Agent):
        return self.eq_attrs(a, b, ['name', 'platform', 'info'])

    def var(self, a:Var, b:Var):
        if self.mapping.get(a):
            return self.mapping[a] is b
        elif self.eq_attrs(a, b, ['value']):
            self.mapping[a] = b
            return True
        return False

    def action(self, a:Action, b:Action):
        eq = self.eq_attrs(a, b, ['action', 'entity', 'bindings', 'types'])
        if a.action == b.action and a.action == ActionType.Call and a.entity.isa(Builtins.Infix):
            expected_name = cast(Type, a.entity.value).name
            actual_name = cast(Type, b.entity.value).name
            if expected_name in OP_COMMUTATIVE or OP_INVERSES.get(expected_name) == actual_name:
                eq = self.eq_set(a.bindings, b.bindings)
        if not eq:
            self.mismatches.append((a, b))
        return eq

#--------------------------------------------------
# Runner
#--------------------------------------------------

def batch(size, ix = 0):
    debugging.DEBUG = False
    gen = Gen(1)
    stats = GenStats()

    for x in range(size):
        gen.test_task(x + ix, stats)

    if DEBUG:
        stats.print_failures()
        stats.print_debugs()

    return stats.finish()

def check_seed(seed):
    return batch(1, seed)

def batches(total=10000, num_processes = None):
    import multiprocessing
    multiprocessing.freeze_support()

    num_processes = num_processes or multiprocessing.cpu_count()  # Or any desired number
    size = total // num_processes

    start = time.perf_counter()
    with multiprocessing.Pool(processes=num_processes) as pool:
        results = pool.starmap(batch, [(size, i * size) for i in range(num_processes)])

    return (time.perf_counter() - start, results)



from abc import ABC
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Sequence

from relationalai.metamodel import Behavior, Var
from relationalai import metamodel as mm

def implements(subtype: mm.Type, supertype: mm.Type, seen: set[mm.Type]|None = None) -> bool:
    if subtype == supertype:
        return True

    if not seen:
        seen = set()
    elif subtype in seen:
        return False

    seen.add(subtype)
    for parent in subtype.parents:
        if implements(parent, supertype, seen):
            return True

    return False

id = 0
def next_id():
    global id
    id += 1
    return id

PRIMITIVE_TYPES = [t for t in vars(mm.Builtins).values() if isinstance(t, mm.Type) and implements(t, mm.Builtins.Primitive)]

@dataclass
class Constant:
    type: mm.Type
    value: Any

    def __hash__(self):
        return hash(self.value)

    def __str__(self):
        return repr(self.value)

Value = Constant | Var


class Provider(ABC):
    def __init__(self, provides: Sequence[mm.Var]):
        self.provides = provides

    def __str__(self):
        default_name = "var"
        return f"Provide({', '.join(f'{v.name or default_name}_{v.id}' for v in self.provides)})"

class EntityProvider(Provider):
    def __init__(self, entity: mm.Var, properties: Sequence[mm.Property]):
        self.entity = entity
        self.properties = {prop: mm.Var(prop.type, f"{entity.name}{entity.id}.{prop.name}") for prop in properties}
        super().__init__([self.entity, *self.properties.values()])

class ComputedProvider(Provider):
    def __init__(self, call: 'Call'):
        self.call = call
        results = [arg for prop, arg in zip(call.op.properties, call.args) if not prop.is_input and isinstance(arg, mm.Var)]
        super().__init__(results)

def fmt_arg(arg: Any):
    match arg:
        case list():
            return f"[{', '.join(fmt_arg(v) for v in arg)}]"
        case mm.Var():
            return f"{arg.name or 'var'}_{arg.id}"
        case _:
            return str(arg)

@dataclass
class Call(Provider):
    id: int = field(default_factory=next_id, init=False)
    op: mm.Task
    args: 'Sequence[Expr]'

    def __post_init__(self):
        self.provides = [arg for arg, prop in zip(self.args, self.op.properties) if not prop.is_input and isinstance(arg, mm.Var)]

    def __str__(self):
        return f"{self.op.name}({', '.join([f'{str(param)}={fmt_arg(arg)}' for (arg, param) in zip(self.args, self.op.properties)])})"

    def __hash__(self):
        return self.id

    def result(self, type: mm.Type) -> mm.Var:
        for var in self.provides:
            if implements(var.type, type):
                return var

        raise Exception(f"Call provides no result of type {type} in {self}.")


Expr = Value | Call

class AssertOp(Enum):
    Bind = "bind"
    Persist = "persist"
    Unpersist = "unpersist"
    # Construct= "construct"

    def __str__(self):
        return self.name

@dataclass
class AssertProperty:
    id: int = field(default_factory=next_id, init=False)
    op: AssertOp
    entity: Var
    property: mm.Property
    value: Expr

    def __str__(self):
        return f"{str(self.entity)}.{self.op}({str(self.property)} = {str(self.value)})"

    def __hash__(self) -> int:
        return self.id

@dataclass
class AssertType:
    id: int = field(default_factory=next_id, init=False)
    op: AssertOp
    entity: Var
    type: mm.Type
    generated: bool = False

    def __str__(self):
        if self.generated:
            return f"{str(self.entity)} = {self.type.name}.{self.op}()"
        else:
            return f"{str(self.entity)}.{self.op}({self.type.name})"

    def __hash__(self) -> int:
        return self.id


Assert = AssertType | AssertProperty

class Quantifier(Enum):
    pass
    # @FIXME: Others not implemented yet
    # Not = "not"


@dataclass
class Subquery:
    quantifier: Quantifier | None
    task: 'Task'

    def __str__(self):
        lines = [f"{str(self.quantifier)} {str(self.task.behavior)}" ]

        body_lines = []
        body_lines.extend([f"  {str(item)}" for item in self.task.items])
        body = "\n".join(body_lines)

        lines.extend([f"  {line}" for line in body.splitlines()])
        return "\n".join(lines)

Action = Provider | Call | Assert | Subquery

@dataclass
class Task:
    id: int = field(default_factory=next_id, init=False)
    behavior: Behavior
    items: list[Action]

    def __str__(self):
        lines = [str(self.behavior)]
        lines.extend([f"  {str(item)}" for item in self.items])
        return "\n".join(lines)

    def __hash__(self) -> int:
        return self.id



import os

import pytest

from relationalai.debugging import logger
from relationalai.util.tracing_logger import TracingLogger
from tests.util import engine_config_fixture, root_span_fixture, snowflake_logger_fixture

# Engine is actually created in end-to-end/action.yml
# keep this in sync with that
@pytest.fixture(scope="session")
def engine_size():
    return "XS"

@pytest.fixture(scope="session")
def engine_config(engine_size):
    yield from engine_config_fixture(engine_size)

# Create a root span that all tests will be a part of
# add `snowflake_logger` as a parameter, to make sure it is initialized before the root span
# and finishes afterwards
@pytest.fixture(scope="session", autouse=True)
def root_span(snowflake_logger):
    yield from root_span_fixture()

# Attach Snowflake logger; shutting down that the end of the session
@pytest.fixture(scope="session", autouse=True)
def snowflake_logger():
    yield from snowflake_logger_fixture()

if os.getenv('ENABLE_TRACING_LOGGER') is not None:
    logger.addHandler(TracingLogger())



import pexpect
import subprocess
import os
import shutil
import tomlkit

import time

from common import (
    hello_world_pyrel_example,
    hello_world_expected_output,
    setup_tempdir,
    example_with_profile,
    example_with_explicit_config,
)

down_arrow = "\033[B"
enter = "\r"
delete = "\x7f"

def test_basic_flow():
    tempdir = setup_tempdir()

    fake_home_dir = os.path.join(tempdir, "home")
    env = {
        "HOME": fake_home_dir,
        "TERM": "dumb",
    }

    print("Writing to example.py...")
    with open("example.py", "w") as f:
        f.write(hello_world_pyrel_example)

    print("Writing to example_with_profile.py...")
    with open("example_with_profile.py", "w") as f:
        f.write(example_with_profile)

    print("Writing to example_with_explicit_config.py...")
    with open("example_with_explicit_config.py", "w") as f:
        f.write(example_with_explicit_config)
    
    print("Writing to example_with_profile_reference.py...")
    with open("example_with_profile_reference.py", "w") as f:
        f.write(hello_world_pyrel_example)

    with open("init.log", "wb") as f:
        print("Init flow...")
        os.environ["TERM"] = "dumb"
        client_id = os.getenv("RAI_CLIENT_ID") or ""
        secret = os.getenv("RAI_CLIENT_SECRET") or ""
        if not client_id or not secret:
            raise ValueError(
                "Environment variables `RAI_CLIENT_ID` and `RAI_CLIENT_SECRET` must be set for this test to be properly run."
            )
        child = pexpect.spawn(".venv/bin/rai init", env=env)  # type: ignore
        child.logfile = f
        child.expect(r"Host platform")
        print("Selecting platform...")
        child.send(down_arrow)
        time.sleep(1)
        child.send(enter)
        child.expect(r"Client ID")
        print(f"Entering Client ID... ({client_id[:5]}...)")
        child.sendline(client_id)
        child.expect(r"Client Secret")
        print(f"Entering Client Secret... ({secret[:5]}...)")
        child.sendline(secret)
        print("Waiting for the engine selection list...")
        child.expect(r"Select an engine")
        print("Selecting engine...")
        time.sleep(0.2)
        child.send(down_arrow)
        child.send(enter)
        child.expect(pexpect.EOF)
        child.close()

    for file in [
        "example.py",
        "example_with_profile.py",
        "example_with_explicit_config.py",
        "example_with_profile_reference.py",
    ]:
        if file == "example_with_profile.py":
            # change the name of the table in raiconfig.toml from default to test:
            with open("raiconfig.toml", "r") as f:
                original = tomlkit.load(f)
                table = tomlkit.table()
                for k, v in original.items():
                    table[k] = v
                doc = tomlkit.document()
                doc.add("profile", tomlkit.table())
                doc["profile"].add("test", table) # type: ignore
            with open("raiconfig.toml", "w") as f:
                f.write(doc.as_string())
        if file == "example_with_profile_reference.py":
            # change the name of the table in raiconfig.toml from default to test:
            with open("raiconfig.toml", "r") as f:
                # content = f.read()
                original = tomlkit.load(f)
                doc = tomlkit.document()
                doc["active_profile"] = "test"
                doc.add("profile", tomlkit.table())
                doc["profile"].add("test", original["profile"]["test"]) # type: ignore
            with open("raiconfig.toml", "w") as f:
                f.write(doc.as_string())
        print(f"Running {file}...")
        result = subprocess.run(
            [".venv/bin/python", file], capture_output=True, text=True, env=env
        )
        assert result.stdout.strip() == hello_world_expected_output.strip()

    shutil.rmtree(tempdir)  # Remove the directory and all its contents



#pyright: reportUnusedExpression=false
import relationalai as rai
from relationalai.clients import config as cfg

import snowflake.connector

config = cfg.Config()

connection = snowflake.connector.connect(
    user=config.get('user'),
    password=config.get('password'),
    account=config.get('account'),
    warehouse=config.get('warehouse', ""),
    role=config.get('role', ""),
    client_store_temporary_credential=True,
    client_request_mfa_token=True,
)

model = rai.Model("MyCoolDatabase", connection=connection)
Person = model.Type("Person")
Adult = model.Type("Adult")

with model.rule():
    Person.add(name="Joe", age=74)
    Person.add(name="Bob", age=40)
    Person.add(name="Jane", age=10)

with model.rule():
    p = Person()
    p.age >= 18
    p.set(Adult)

with model.query() as select:
    a = Adult()
    z = select(a, a.name, a.age)

print("People!\n")
print(z.results)










#pyright: reportUnusedExpression=false
from typing import Tuple
import rich
import relationalai as rai
from relationalai.std import aggregates
from relationalai.std.graphs import Graph

model = rai.Model("SFPagerank")

#--------------------------------------------------
# Initialize Snowflake data
#--------------------------------------------------

Account = model.Type("Account", source="sandbox.public.accounts")

Transaction = model.Type("Transaction", source="sandbox.public.transactions")
Transaction.define(
    from_ = (Account, "from_account", "account_id"),
    to = (Account, "to_account", "account_id")
)

#--------------------------------------------------
# Add types
#--------------------------------------------------

Merchant = model.Type("Merchant")
Person = model.Type("Person")

with model.rule():
    a = Account()
    with a.account_type == "Merchant":
        a.set(Merchant)
    with a.account_type == "User":
        a.set(Person)

#--------------------------------------------------
# Create a graph
#--------------------------------------------------

graph = Graph(model)
Node, Edge = graph.Node, graph.Edge

Node.extend(Account, name=Account.name)
Node.extend(Person, kind="person")
Node.extend(Merchant, kind="merchant", label=Merchant.name)

with model.rule():
    t = Transaction()
    Edge.add(t.from_, t.to, amount=t.amount)

with model.rule():
    m = Merchant()
    rank = graph.compute.pagerank(m)
    Node(m).set(rank=rank)
    m.set(rank=rank)

#--------------------------------------------------
# Visualize the graph
#--------------------------------------------------

graph.visualize(
    three=True,
    node_label_size_factor=1.2,
    use_links_force=True,
    links_force_distance=140,
    node_hover_neighborhood=True,
    style={
        "node": {
            "size": lambda n: (n.get("rank", 0.01) * 100) ** 2.2 + 10,
            "color": lambda n: "green" if n["kind"] == "merchant" else "#bbb",
            "hover": lambda n: f'{n["name"]} ({n.get("rank", 0):,.3f})' if n.get("rank") else n["name"]
        },
        "edge": {
            "opacity": 0.3,
            "color": "#ccc",
            "size": lambda t: t["amount"] / 50
        }
    }
).display()

#--------------------------------------------------
# Export the analysis
#--------------------------------------------------

@model.export("sandbox.public")
def merchant_rank(minimum: float) -> Tuple[str, float, int]:
    m = Merchant()
    m.rank >= minimum
    t = Transaction(to=m)
    total = aggregates.sum(t, t.amount, per=[m])
    return m.name, m.rank, total #type: ignore

print("\nCalling merchant rank from Snowflake!\n")
for row in model.resources._exec("call sandbox.public.merchant_rank(0.04);"):
    rich.print(row)



import time
from hypothesis import Verbosity, given, settings
from gentest.gen.action import gen_task_body
from gentest.gen.scope import GenScope, GenScopeBuilder
from gentest.gen.task import gen_task
from relationalai import metamodel as mm
from gentest.gen import ir
from gentest.gen.context import fixture
from gentest import fixtures



if __name__ == "__main__":
    ctx = fixture(fixtures.person_place_thing).finish()
    print(str(ctx))

    root_builder = GenScopeBuilder(ctx)

    root = root_builder.finish()
    # example_scope = gen_provision_stage(root).example()
    # print(str(example_scope[1]))

    runs = 0

    @settings(max_examples=100, verbosity=Verbosity.quiet)
    # @given(construct_scope(root).flatmap(gen_expr_examples))
    # @given(gen_expr_examples(example_scope))
    @given(gen_task_body(root, 5))
    def show(stuff: tuple[list[ir.Action], GenScope]):
        global runs
        runs += 1
        print("-"*80)
        actions, _ = stuff
        print("    -", "\n    - ".join(str(v) for v in actions))
        # scope, values = stuff
        # for v in values:
        #     if isinstance(v, ir.Call):
        #         break
        # else:
        #     return

        # print(str(scope))
        # print("Values")
        # for v in values:
        #     rich.print(f"    - {str(v)}")

    action_count = 0
    best_score = 0
    best = None

    @settings(max_examples=100, verbosity=Verbosity.quiet)
    @given(gen_task(root))
    def show_task(task: mm.Task):
        global runs, action_count, best_score, best
        runs += 1
        action_count += len(task.items)

        score = len(task.items)
        if score > best_score:
            best_score = score
            best = task

        # print("task", "-"*75)
        # print(str(task))
        # print("="*80)

    start = time.time_ns()
    show_task()
    end = time.time_ns()
    elapsed = (end - start) / 1000000
    print(f"Completed {runs} runs in {elapsed:.2f}ms. Avg { elapsed / runs:.4f}ms per task.")
    print(f"with {action_count} total actions Avg {action_count / runs:.1f} actions per task, { elapsed / action_count:.4f}ms per action in task.")

    print("Ex.")
    print(str(best))



import relationalai as rai

model = rai.Model(name=globals().get("name", "test_first"), config=globals().get("config"))

Person = model.Type("Person")

with model.rule():
    Person.add(name="Sam Watson")
    Person.add(name="Pete Vilter")
    Person.add(name="Josh Cole")

with model.query() as select:
    p = Person()
    z = select(
        p.name
    )  # @NOTE: we don't actually need to use the result, it'll get snapshotted all the same

print(z.results)



import gravis as gv
import relationalai as rai

model = rai.Model("ibdknox")

Person = model.Type("Person")
Adult = model.Type("Adult")
Customer = model.Type("Customer")
Criminal = model.Type("Criminal")
Transaction = model.Type("Transaction")
Suspicious = model.Type("Suspicious")
ImmediateReview = model.Type("ImmediateReview")

with model.rule():
    (id, name, age, account, criminal) = rai.csv.load_file(model, "data/people.csv")
    p = Person.add(name=name, age=age, csv_id=id)
    with account:
        p.set(Customer, account=account)
    with criminal:
        p.set(Criminal)

with model.rule():
    (txn_id, date, from_id, to_id, amount) = rai.csv.load_file(model, "data/transactions.csv")
    Transaction.add(txn_id=txn_id, date=date, from_=Customer(csv_id=from_id), to_=Customer(csv_id=to_id), amount=amount)

with model.query() as select:
    s = Transaction()
    res = select(s, s.from_.name, s.to_.name, s.amount)

with model.rule():
    t = Transaction(from_=Customer(), to_=Criminal())
    t.set(Suspicious)
    with t.amount > 40000:
        t.set(ImmediateReview, review_reason="Large transaction to criminal")

with model.query() as select:
    s = ImmediateReview()
    res = select(s.from_.name, s.to_.name, s.amount)

with model.rule():
    s = Suspicious(Transaction)
    s.from_.set(Suspicious)
    s.to_.set(Suspicious)

with model.query() as select:
    c = Criminal()
    criminals = select(c)

criminals = set(criminals.results['criminal'])

# Create edges

with model.query() as select:
    s = Suspicious(Transaction)
    edges = select(s.from_, s.to_, s.amount)

# Create nodes

with model.query() as select:
    s = Suspicious(Person)
    nodes = select(s, s.name)

graph1 = {
    'graph': {
        'directed': True,
        'metadata': {
            'arrow_size': 5,
            'background_color': 'black',
            'edge_size': 3,
            'edge_label_size': 10,
            'edge_label_color': 'white',
            'edge_color': '#777',
            'node_size': 15,
            'node_color': 'white',
            'node_label_color': 'white',
            'node_label_size': 10,
        },
        'nodes': {
            node_id: {'metadata': {'name': name,
                                  'color': '#f88' if node_id in criminals else 'white',
                                  }} for (node_id, name) in nodes
        },
        'edges': [
            {
                'source': source,
                'target': target,
                'metadata': {
                    'color': '#f44' if amount > 40000 else '#777',
                             'label_color': 'orange' if amount > 40000 else '#ccc',
                            'amount': f"${amount/1000:,.0f}k"}
            } for (source, target, amount) in edges
        ],
    }
}

fig = gv.vis(graph1,
       # show_edge_label=True,
       node_label_data_source='name',
       edge_label_data_source='amount',
       layout_algorithm='forceAtlas2Based',
       central_gravity=2.69,
       spring_length=13)
fig.display()




import relationalai as rai
from relationalai import std

# Query snapshots are output in same order as these lists.
NO_ARGS = [
    "num_triangles",
    "triangles",
]

ONE_ARG = [
    "num_triangles", # included here because of optional 'node' parameter
    "triangles", # included here because of optional 'node' parameter
]

THREE_ARGS = [
    "is_triangle",
]

model = rai.Model(name=globals().get("name", "test_triangles"), config=globals().get("config"))
Object = model.Type("Object")
Relationship = model.Type("Relationship")

# Create a graph with 4 nodes and 4 edges, forming a triangle with a dangling node.
with model.rule(dynamic=True):
    object = [Object.add(id=i) for i in range(4)]
    for i in range(3):
        Relationship.add(from_=object[i], to=object[(i + 1) % 3])
    Relationship.add(from_=object[0], to=object[3])

graph = std.graphs.Graph(model, undirected=True)

with model.rule():
    r = Relationship()
    graph.Edge.add(r.from_, r.to)

for algo_name in NO_ARGS:
    with model.query(dynamic=True) as select:
        algo = getattr(graph.compute, algo_name)()
        if isinstance(algo, tuple):
            select(*algo)
        else:
            select(algo)

for algo_name in ONE_ARG:
    with model.query(dynamic=True) as select:
        o = Object()
        algo = getattr(graph.compute, algo_name)(o)
        if isinstance(algo, tuple):
            select(o.id, *algo)
        else:
            select(o.id, algo)

for algo_name in THREE_ARGS:
    with model.query(dynamic=True) as select:
        o1, o2, o3 = Object(), Object(), Object()
        o1.id < o2.id < o3.id
        algo = getattr(graph.compute, algo_name)(o1, o2, o3)
        if isinstance(algo, tuple):
            select(o1.id, o2.id, o3.id, *algo)
        else:
            select(o1.id, o2.id, o3.id, algo)



import relationalai as rai

graph = rai.Model("RecursiveExample")
Edge = graph.Type("Edge")

## Find all paths reachable from a given start node and an initial set of 1-hop edges.

with graph.rule():
    # Create some initial edges
    Edge.add(start=1, finish=2)
    Edge.add(start=2, finish=3)
    Edge.add(start=3, finish=5)
    Edge.add(start=4, finish=6)
    # Expecting 1 -> 2 -> 3 -> 5, 4 -> 6

with graph.rule():
    first = Edge()
    second = Edge(start = first.finish)
    Edge.add(start=first.start, finish=second.finish)

with graph.query() as select:
    edge = Edge()
    response = select(edge, edge.start, edge.finish)

print(response)



from typing import TypeVar
import threading
from hypothesis.strategies._internal.strategies import SearchStrategy
from hypothesis.strategies._internal.recursive import LimitReached

T = TypeVar("T")

class GroupLimitedStrategy(SearchStrategy):
    """A strategy which, when called within a GroupLimiterStrategy, will only run its base strategy if the limit hasn't been reached. When the limit is reached, it'll call the optional exhausted strategy if provided, or throw."""
    _active_group_limits = threading.local()

    @classmethod
    def limit_group(cls, group: str, limit: int|None):
        if limit is None:
            delattr(cls._active_group_limits, group)
        else:
            setattr(cls._active_group_limits, group, limit)

    @classmethod
    def current_limit(cls, group: str):
        return getattr(cls._active_group_limits, group, None)

    @classmethod
    def is_limited(cls, group: str):
        return getattr(cls._active_group_limits, group, None) is not None

    def __init__(self, group: str, strategy: SearchStrategy, exhausted: SearchStrategy|None):
        super().__init__()
        self.group = group
        self.base_strategy = strategy
        self.exhausted = exhausted

    @property
    def remaining(self):
        return getattr(GroupLimitedStrategy._active_group_limits, self.group, None)

    @remaining.setter
    def remaining(self, value):
        return setattr(GroupLimitedStrategy._active_group_limits, self.group, value)

    def do_validate(self):
        self.base_strategy.validate()

    def do_draw(self, data):
        remaining = self.remaining
        if remaining is None:
            raise Exception(f"GroupLimitedStrategy for group {self.group} can only be drawn from within an active group limit. Make sure you've wrapped your draws in a `with group_limit('{self.group}', NUM):`")

        if remaining <= 0:
            if self.exhausted is not None:
                return data.draw(self.exhausted)
            else:
                raise LimitReached
        self.remaining = remaining - 1
        return data.draw(self.base_strategy)

    def __repr__(self):
        return f"GroupLimitedStrategy({self.group}, {self.base_strategy!r})"

class GroupLimiterStrategy(SearchStrategy):
    """A strategy which constrains all descendant GroupLimitedStrategy's of the given groups with the specified limit."""
    def __init__(self, inner: SearchStrategy, limits: dict[str, int]):
        super().__init__()
        self.inner_strategy = inner
        self.limits = limits

    def do_validate(self):
        self.inner_strategy.validate()

    def do_draw(self, data):
        owned_limits = []
        for group, limit in self.limits.items():
            if not GroupLimitedStrategy.is_limited(group):
                owned_limits.append(group)
                GroupLimitedStrategy.limit_group(group, limit)

        try:
            return data.draw(self.inner_strategy)
        finally:
            for group in owned_limits:
                GroupLimitedStrategy.limit_group(group, None)

    def __repr__(self):
        return f"GroupLimiterStrategy({self.limits}, {self.inner_strategy!r})"


def limited_by_group(group: str, strategy: SearchStrategy[T], exhausted: SearchStrategy|None) -> SearchStrategy[T]:
    """Wrap a strategy in a limiter. When `group` has an active limit (set using the `limit_group()` ctx manager), strategy will bail when the limit runs out."""
    return GroupLimitedStrategy(group, strategy, exhausted)

def group_limiter(strategy: SearchStrategy[T], **limits: int) -> SearchStrategy[T]:
    return GroupLimiterStrategy(strategy, limits)














# import threading
# from contextlib import contextmanager

# from hypothesis.errors import InvalidArgument
# from hypothesis.internal.reflection import get_pretty_function_description
# from hypothesis.internal.validation import check_type
# from hypothesis.strategies._internal.strategies import (
#     OneOfStrategy,
#     SearchStrategy,
#     check_strategy,
# )
# from hypothesis.strategies._internal.recursive import LimitReached


# class GroupLimitedStrategy(SearchStrategy):
#     _active_group_limits = threading.local()

#     @classmethod
#     def limit_group(cls, group: str, limit: int|None):
#         if limit == None:
#             delattr(cls._active_group_limits, group)
#         else:
#             setattr(cls._active_group_limits, group, limit)

#     @classmethod
#     def current_limit(cls, group: str):
#         return getattr(cls._active_group_limits, group, None)

#     def __init__(self, group: str, strategy: SearchStrategy, require_limit = False):
#         super().__init__()
#         self.group = group
#         self.base_strategy = strategy
#         self.require_limit = require_limit

#     @property
#     def remaining(self):
#         return getattr(GroupLimitedStrategy._active_group_limits, self.group, None)

#     @remaining.setter
#     def remaining(self, value):
#         return setattr(GroupLimitedStrategy._active_group_limits, self.group, value)

#     def do_validate(self):
#         self.base_strategy.validate()

#     def do_draw(self, data):
#         remaining = self.remaining
#         if remaining == None:
#             assert not self.require_limit
#             return data.draw(self.base_strategy)

#         if remaining <= 0:
#             raise LimitReached
#         self.remaining = remaining - 1
#         return data.draw(self.base_strategy)

#     def __repr__(self):
#         return f"GroupLimitedStrategy({self.group}, {self.base_strategy!r})"

# @contextmanager
# def limit_group(group: str, limit: int):
#     prev = GroupLimitedStrategy.current_limit(group)
#     GroupLimitedStrategy.limit_group(group, limit)
#     try:
#         yield
#     finally:
#         GroupLimitedStrategy.limit_group(group, prev)

# def limited_group(group: str, strategy: SearchStrategy, require_limit = False):
#     """Wrap a strategy in a limiter. When `group` has an active limit (set using the `limit_group()` ctx manager), strategy will bail when the limit runs out."""



from typing import List, TypeVar, cast
from .metamodel import Behavior, Type, Property, Task, Var, Action, Agent, next_id, AllItems
import copy
from . import debugging
import time


#--------------------------------------------------
# Compiler Pass
#--------------------------------------------------

T = TypeVar('T', bound='AllItems')

class Pass:
    backlinks = {}
    forwardlinks = {}

    def __init__(self, copying = True) -> None:
        self.seen = set()
        self.copying = copying

    def clone(self, item:T) -> T:
        copied = copy.copy(item)
        copied.id = next_id()
        Pass.backlinks[copied] = item
        Pass.forwardlinks[item] = copied
        return copied

    def walk(self, item: T, parent=None) -> T:
        if not item or item in self.seen:
            return Pass.forwardlinks.get(item, item)
        if self.copying and not Pass.backlinks.get(item) and not isinstance(item, Var) and (isinstance(item, Task) or not isinstance(item, Type)):
            item = self.clone(item)
        elif self.copying and isinstance(item, Var) and isinstance(item.value, Task) and len(item.value.items):
            item = self.clone(item)
        self.seen.add(item)

        if isinstance(item, Task):
            self.task(cast(Task, item), parent)
        elif isinstance(item, Type):
            self.type(item, parent)
        elif isinstance(item, Property):
            self.property(item, parent)
        elif isinstance(item, Agent):
            self.agent(item, parent)
        elif isinstance(item, Var):
            self.var(item, parent)
        elif isinstance(item, Action):
            self.action(item, parent)

        return item

    def walk_all(self, items, parent=None):
        if self.copying:
            return [self.walk(item, parent) for item in items]
        else:
            for item in items:
                self.walk(item, parent)
            return items

    def reset(self):
        self.seen.clear()

    def type(self, type: Type, parent=None):
        type.properties = self.walk_all(type.properties, type)

    def property(self, property: Property, parent=None):
        property.type = self.walk(property.type)

    def task(self, task: Task, parent=None):
        # task.properties = self.walk_all(task.properties, task)
        if task.agent:
            task.agent = self.walk(task.agent)
        if task.behavior == Behavior.Sequence:
            self.sequence(task, parent)
        elif task.behavior == Behavior.Query:
            self.query(task, parent)
        elif task.behavior == Behavior.Union:
            self.union(task, parent)
        elif task.behavior == Behavior.OrderedChoice:
            self.ordered_choice(task, parent)
        elif task.behavior == Behavior.Catch:
            self.catch(task, parent)

    def sequence(self, seq: Task, parent=None):
        seq.items = self.walk_all(seq.items, seq)

    def query(self, query: Task, parent=None):
        query.items = self.walk_all(query.items, query)

    def union(self, union: Task, parent=None):
        union.items = self.walk_all(union.items, union)

    def ordered_choice(self, ordered_choice: Task, parent=None):
        ordered_choice.items = self.walk_all(ordered_choice.items, ordered_choice)

    def catch(self, catch: Task, parent=None):
        catch.items = self.walk_all(catch.items, catch)

    def agent(self, agent: Agent, parent=None):
        pass

    def var(self, var: Var, parent=None):
        var.type = self.walk(var.type, parent)
        if (isinstance(var.value, Type) or isinstance(var.value, Property)) \
            and (not isinstance(var.value, Task) or len(var.value.items)):
            var.value = self.walk(var.value, parent)

    def action(self, action: Action, parent=None):
        action.entity = self.walk(action.entity, parent)
        action.types = self.walk_all(action.types, parent)
        action.bindings = {k: self.walk(v, parent) for k, v in action.bindings.items()}

#--------------------------------------------------
# Emitter
#--------------------------------------------------

class Emitter:
    def emit(self, task: Task|Var) -> str:
        return ""

#--------------------------------------------------
# Compilation
#--------------------------------------------------

class Compilation:
    def __init__(self, task:Task, debug=debugging.DEBUG):
        self.task = task
        self.passes = []
        self.emitted = ""
        self.emit_time = 0.0
        self.debug = debug

    def pass_result(self, pass_:Pass, rewritten_task:Task, elapsed:float):
        if self.debug:
            self.passes.append((pass_.__class__.__name__, str(rewritten_task), elapsed))

    def get_source(self):
        pyrel_info = debugging.get_source(self.task)
        if not pyrel_info:
            sources = []
            for i in self.task.items:
                source = debugging.get_source(i)
                if source:
                    if not pyrel_info:
                        pyrel_info = source
                    if source.source not in sources:
                        sources.append(source.source)
            pyrel_block = "\n".join(sources)
        else:
            pyrel_block = pyrel_info.source
        if not pyrel_info:
            pyrel_info = debugging.SourceInfo()
        return (pyrel_info.file, pyrel_info.line, pyrel_block)

#--------------------------------------------------
# Compiler
#--------------------------------------------------

class Compiler:
    def __init__(self, emitter:Emitter, passes:List[Pass] = []):
        self.emitter = emitter
        self.passes = passes

    def rewrite(self, task: Task, compilation:Compilation):
        for pass_ in self.passes:
            start = time.perf_counter()
            pass_.reset()
            task = pass_.walk(task)
            compilation.pass_result(pass_, task, time.perf_counter() - start)
        return task

    def compile(self, task: Task):
        compilation = Compilation(task)
        task = self.rewrite(task, compilation)
        start = time.perf_counter()
        self.emitter.reset()
        code = self.emitter.emit(task)
        compilation.emitted = code
        compilation.emit_time = time.perf_counter() - start
        debugging.handle_compilation(compilation)
        return code



from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, cast
from relationalai.analysis.mechanistic import Mechanism

from pandas import DataFrame
from ..clients.client import Client
from .. import debugging
from .. import dsl, rel, metamodel as m

@dataclass
class Query:
    task: m.Task
    ix: int
    code: str|None = None
    err: Exception|None = None
    result: Any = None
    mech: Mechanism|None = None

    def to_json(self):
        return {
            "task": str(self.task),
            "ix": self.ix,
            "code": self.code,
            "err": str(self.err),
            "mech": self.mech.stack[-1] if self.mech else None
        }


@dataclass
class Install:
    task: m.Task
    name: str|None = None
    code: str|None = None
    err: Exception|None = None
    mech: Mechanism|None = None

    def to_json(self):
        return {
            "task": str(self.task),
            "name": self.name,
            "code": self.code,
            "err": str(self.err),
            "mech": self.mech.stack[-1] if self.mech else None
        }


class Document():
    blocks: list[Query | Install]
    query_count: int

    def __init__(self):
        self.blocks = []
        self.query_count = 0

    def __str__(self):
        parts = []
        h = "==[ Document ]"
        print(f"{h}{'='*(80 - len(h))}")
        for block in self.blocks:
            match block:
                case Query():
                    h = f"----[ Query {block.ix} ]"
                    parts.append(f"{h}{'-'*(80 - len(h))}")
                    parts.append(str(block.task))
                case Install():
                    h = f"----[ Rule {block.name} ]"
                    parts.append(f"{h}{'-'*(80 - len(h))}")
                    parts.append(str(block.task))

        return "\n".join(parts)

class Config():
    def get(self, key:str, default:Any=None):
        return default

class NoopClient():
    def __init__(self, *_args, config=None, **_kwargs):
        self.compiler = rel.Compiler(config or Config())
        self.resources = Resources(config or Config())

    def query(self, task: m.Task, *args, **kwargs):
        self.compiler.compile(task)
        return DataFrame()

    def install(self, name: str, task: m.Task, *args, **kwargs):
        self.compiler.compile(task)

    def export_udf(self, name, inputs, out_fields, task):
        pass

    def exec_raw(self, raw_code:str):
        raise Exception("Cannot exec raw code in test executor")

    def load_raw_file(self, path:str):
        raise Exception("Cannot load raw file in test executor")

proxied_clients: list[Document] = []

class Resources():
    def __init__(self, config:Config):
        self.config = config

def proxy_client(client_class, throw_on_error = True):
    class ProxiedClient(client_class, Document):
        def __init__(self, *args, throw_on_error = True, **kwargs):
            self.blocks: list[Query | Install] = []
            self.query_count = 0
            self.throw_on_error = throw_on_error
            # self.capture_handler = CaptureHandler()

            # debugging.logger.addHandler(self.capture_handler)
            super().__init__(*args, **kwargs)
            proxied_clients.append(self)

        # def __del__(self):
        #     # @FIXME: This is apparently not reliable
        #     debugging.logger.removeHandler(self.capture_handler)

        @contextmanager
        def capture_debugging(self):
            orig_logger_debug = debugging.logger.debug
            msgs = []
            def capture(msg):
                msgs.append(msg)
                orig_logger_debug(msg)

            setattr(debugging.logger, "debug", capture)
            try:
                yield msgs
            finally:
                setattr(debugging.logger, "debug", orig_logger_debug)

        def get_emitted(self, msgs: list):
            msgs[0]["emitted"] if len(msgs) > 0 and msgs[0]["event"] == "compilation" else None

        def query(self, task:m.Task, *args, **kwargs):
            ix = self.query_count
            self.query_count += 1
            res = None
            with self.capture_debugging() as msgs:
                mech = None
                try:
                    mech = Mechanism(task)
                    res = super().query(task, *args, **kwargs)
                    self.blocks.append(Query(task, ix, self.get_emitted(msgs), None, res, mech))

                except Exception as err:
                    self.blocks.append(Query(task, ix, self.get_emitted(msgs), err, None, mech))
                    if self.throw_on_error:
                        raise

            return res

        def install(self, name, task:m.Task, *args, **kwargs):
            with self.capture_debugging() as msgs:
                mech = None
                try:
                    mech = Mechanism(task)
                    super().install(name, task, *args, **kwargs)
                    self.blocks.append(Install(task, name, self.get_emitted(msgs), None, mech))
                except Exception as err:
                    self.blocks.append(Install(task, name, self.get_emitted(msgs), err, mech))
                    if self.throw_on_error:
                        raise

    return ProxiedClient

Executor = proxy_client(NoopClient, throw_on_error=False)

def Graph(name, engine:str|None=None, dry_run=False, config=None):
    client = cast(Client, Executor(name, engine, dry_run, config=config or Config()))
    return dsl.Graph(client, name)



from hypothesis import settings, given, strategies as gen
from hypothesis.internal.conjecture.data import ConjectureData
from gentest.gen.document import gen_document
from relationalai.clients.test import Document

from gentest import fixtures
from gentest.gen.context import fixture
from gentest.gen.scope import GenScope
from gentest.validate.diff import diff_to_str

from gentest.emit import DocumentEmitter
from gentest.validate.mapping import map_document_ids, normalize_expected_task
from gentest.util import fmt_task
from gentest.harness.database import HYPO_DB
from gentest.validate.errors import RoundtripError, EmitError, ExecError, DiffError
from gentest.validate.roundtrip import roundtripper, exec_and_run_callback

# @NOTE: For reasons I don't understand, stack line numbers change between runs when we directly raise here instead of return the errors
#        which triggers the flaky test detection, so this weird pattern of returning them is necessary.
def document_roundtrip(expected: Document, conjecture: ConjectureData) -> RoundtripError|None:
    __tracebackhide__ = True

    for block in expected.blocks:
        normalize_expected_task(block.task)

    try:
        doc = DocumentEmitter.from_document(expected)
        code = doc.stringify().strip()
    except Exception as err:
        # @NOTE: unfortunately including these tracebacks incorrectly triggers Hypothesis's flaky test detection, because
        # it seems to expect the exact same error to be returned from the same position every shrink.
        # This doesn't make any sense to me, but I don't have more time to investigate it right now.
        # tb = sys.exc_info()[2]
        return EmitError(expected, conjecture.buffer, err) # .with_traceback(tb)

    try:
        actual: Document = exec_and_run_callback(code)
    except Exception as err:
        # tb = sys.exc_info()[2]
        return ExecError(expected, conjecture.buffer, code, err) # .with_traceback(tb)

    zipped_blocks = map_document_ids(expected, actual)

    for (expected_block, actual_block) in zipped_blocks:
        expected_str = fmt_task(expected_block.task)
        actual_str = fmt_task(actual_block.task)
        if expected_str != actual_str:
            # @FIXME: we're accidentally erasing all the information about the document itself here.
            #         Even though they're very similar this should probably be a different error type
            #         unless I can think of a good way to handle that attribution?
            return DiffError(expected, conjecture.buffer, code, actual, diff_to_str(expected_str, actual_str))

ctx = fixture(fixtures.person_place_thing).finish()
root = GenScope(ctx)

@roundtripper
# @seed(3) # reproduces flaky test results
# @reproduce_failure("6.92.0", b'AXicY2RiYGBkYGRkQAAAAHIABw==') # Repros two vars ground by same ent, prop
@settings(max_examples=50, database=HYPO_DB, )
@given(gen_document(root), gen.data())
def test_document_roundtrip(document: Document, data: gen.DataObject):
    err = document_roundtrip(document, data.conjecture_data)
    if err:
        raise err



import pytest
import relationalai as rai
from relationalai import std
from relationalai.errors import RAIException

# Query snapshots are output in same order as this list.
ALGOS = [
    "louvain",
    "weakly_connected_component",
    "triangle_community",  # Expects empty dataframe b/c graph has no triangles.
    "infomap",
    "label_propagation",
]

model = rai.Model(name=globals().get("name", "test_community"), config=globals().get("config"))
Object = model.Type("Object")
Relationship = model.Type("Relationship")

with model.rule():
    obj1 = Object.add(id=1)
    obj2 = Object.add(id=2)
    obj3 = Object.add(id=3)
    Relationship.add(from_=obj1, to=obj2)

graph = std.graphs.Graph(model, undirected=True)
graph.Node.extend(Object)

with model.rule():
    r = Relationship()
    graph.Edge.add(r.from_, r.to)

for algo_name in ALGOS:
    with model.query(tag=algo_name) as select:
        o = Object()
        algo = getattr(graph.compute, algo_name)(o)
        select(o, algo)


# Check that exceptions are raised when invalid parameters are passed.
PARAMETERS = {
    "infomap": {
        "max_levels": [-1, 0, -1.23, 0.0, 1.23, "a", []],
        "max_sweeps": [-1, -1.23, 0.0, 1.23, "a", []],
        "level_tolerance": [-1.0, -1, "a", []],
        "sweep_tolerance": [-1.0, -1, "a", []],
        "teleportation_rate": [-1.0, -1, 0.0, 0, 1.0001, 2, "a", []],
        "visit_rate_tolerance": [-1.0, -1, 0.0, 0, "a", []],
        "randomization_seed": [-1, 1.23, "a", []],
    },
    "label_propagation": {
        "max_sweeps": [-1, -1.23, 0.0, 1.23, "a", []],
        "randomization_seed": [-1, 1.23, "a", []],
    },
    "louvain": {
        "max_levels": [-1, 0, -1.23, 0.0, 1.23, "a", []],
        "max_sweeps": [-1, -1.23, 0.0, 1.23, "a", []],
        "level_tolerance": [-1.0, -1, "a", []],
        "sweep_tolerance": [-1.0, -1, "a", []],
        "randomization_seed": [-1, 1.23, "a", []],
    },
}
with model.query(dynamic=True):
    o = Object()
    for algo_name, params in PARAMETERS.items():
        for param_name, values in params.items():
            for value in values:
                with pytest.raises(RAIException):
                    getattr(graph.compute, algo_name)(o, **{param_name: value})



import relationalai as rai

def main():
    """
    This example shows how to issue SQL queries directly
    using the same connection that PyRel uses.
    """
    model = rai.Model("Example")

    # one-line example:
    print(model.resources._exec("select 1 + 1;").fetch_pandas_all())

    # accessing the connection object directly:
    conn = model.resources._conn
    with conn.cursor() as cursor:
        cursor.execute("select 2 + 2;")
        print(cursor.fetch_pandas_all())

main()



import relationalai as rai
from relationalai import std


model = rai.Model(name=globals().get("name", "test_strings"), config=globals().get("config"))
Person = model.Type("Person")
LikePattern = model.Type("LikePattern")

with model.rule():
    Person.add(first="John", last="Coltrane")
    Person.add(first="Miles", last="Davis")
    Person.add(name="Herbie Hancock")
    # Used in the `like` test, which relies on there being exactly one LikePattern instance.
    LikePattern.add(pattern="Mi%")

# Check that 'length' works.
with model.query(tag="length") as select:
    p = Person()
    f_length = std.strings.length(p.last)
    v_length = std.strings.length("Four")
    select(p, p.last, f_length, v_length)

# Check that 'contains' works.
with model.query(tag="contains") as select:
    p = Person()
    with model.union() as u:
        with model.scope():
            std.strings.contains(p.last, "trane")
            u.add(p)
        with model.scope():
            std.strings.contains("SMiles", p.first)
            u.add(p)
    select(p, p.first, p.last)

# Check that 'ends_with' works.
with model.query(tag="ends_with") as select:
    p = Person()
    with model.union() as u:
        with model.scope():
            std.strings.ends_with(p.last, "trane")
            u.add(p)
        with model.scope():
            std.strings.ends_with("SMiles", p.first)
            u.add(p)
    select(u, u.first, u.last)

# Check that 'lowercase' and 'uppercase' work.
with model.query(tag="lowercase") as select:
    p = Person()
    lower = std.strings.lowercase(p.first)
    upper = std.strings.uppercase(p.first)
    select(p, p.first, lower, upper)

# Check that 'join' works.
with model.query(tag="join") as select:
    p = Person()
    name = std.strings.join([p.first, p.last], " ")
    name2 = std.strings.join(["FIRST", p.last], " ")
    name3 = std.strings.join([p.first, "LAST"], " ")
    sep = std.Vars(1)
    sep == ""
    name4 = std.strings.join(["FIRST", "LAST"], sep)
    select(p, p.first, p.last, name, name2, name3, name4)

# Check that concat works:
with model.query(tag="concat") as select:
    p = Person()
    first_name = std.strings.concat(p.first, "")
    last_name = std.strings.concat("", p.last)
    full_name = std.strings.concat(p.first, p.last)
    new_name = std.strings.concat("HEY", "YOU")
    select(p, first_name, last_name, full_name, new_name)

# Check that replace works.
with model.query(tag="replace") as select:
    p = Person(first="John", last="Coltrane")
    s = Person(first="Miles", last="Davis").first
    new1 = std.strings.replace(p.first, old="ohn", new="ohnny")
    new2 = std.strings.replace(p.first, old="ohn", new=s)
    new3 = std.strings.replace(p.first, old=p.first, new=s)
    select(p, p.first, new1, new2, new3)

# Check that 'like' works.
with model.query(tag="like") as select:
    p = Person()
    with model.union() as u:
        with model.scope():
            std.strings.like(p.last, "Col%")
            u.add(p)
        with model.scope():
            pattern = LikePattern()
            std.strings.like(p.first, pattern.pattern)
            u.add(p)
    select(u, u.first, u.last)

# Check that 'regex_match' and 'regex_compile' work.
with model.query(tag="regex_match_and_compile") as select:
    p = Person()
    pattern = std.strings.regex_compile(r"Col.*")
    std.strings.regex_match(p.last, pattern)
    select(p, p.first, p.last)

# Check that 'split' work
with model.query(tag="split") as select:
    p = Person()
    ix, part = std.strings.split(p.name, " ")
    select(ix, part)

# Check that 'split_part' works.
with model.query(tag="split_part") as select:
    p = Person()
    first = std.strings.split_part(p.name, " ", 0)
    last = std.strings.split_part(p.name, " ", 1)
    select(p, p.name, first, last)



import relationalai as rai
from relationalai import std

# Query snapshots are output in same order as this list.
UNWEIGHTED_ALGOS = [
    "pagerank",
    "eigenvector_centrality",
    "degree_centrality",
    "betweenness_centrality",
]

WEIGHTED_ALGOS = ["weighted_degree_centrality"]

model = rai.Model(name=globals().get("name", "test_centrality"), config=globals().get("config"))
Object = model.Type("Object")
Relationship = model.Type("Relationship")

with model.rule():
    obj1 = Object.add(id=1)
    obj2 = Object.add(id=2)
    Relationship.add(from_=obj1, to=obj2, weight=10)
# eigenvector_centrality only supports undirected graphs.
undirected_graph = std.graphs.Graph(model, undirected=True)

with model.rule():
    r = Relationship()
    undirected_graph.Edge.add(r.from_, r.to)

for algo_name in UNWEIGHTED_ALGOS:
    with model.query(tag=algo_name) as select:
        o = Object()
        algo = getattr(undirected_graph.compute, algo_name)(o)
        select(o, algo)

weighted_graph = std.graphs.Graph(model, undirected=True, weighted=True)

with model.rule():
    r = Relationship()
    weighted_graph.Edge.add(r.from_, r.to, weight=r.weight)

for algo_name in WEIGHTED_ALGOS:
    with model.query(tag=algo_name) as select:
        o = Object()
        algo = getattr(weighted_graph.compute, algo_name)(o)
        select(o, algo)



from .clients import config as cfg
from .errors import InvalidVersionKind, RAIInvalidVersionWarning
from packaging.version import Version

# A simple container for 2 version objects, a lower bound and an upper bound.
#
# Equivalent to a packaging SpecifierSet defined as ">={lower_bound}, <{upper_bound}", but
# has the benefit that we can check whether a version that is not contained by the set is
# lower than the lower bound or higher than the higher bound, which is important for error
# reporting.
def version_range(low:str, high:str) -> tuple[Version, Version]:
    return (Version(low), Version(high))

# A list of rel libraries that the current version of relationalai depends on, including the
# range of compatible versions.
DEPENDENCIES = [
    ('std', '24c1872c-de8a-5b12-7648-3d72f007a7a9', version_range('0.1.0', '0.2.0')),
    ('graphlib', '04cd9c07-16aa-4433-bdd7-adfdfaee317d', version_range('0.1.0', '0.2.0'))
]

#--------------------------------------------------
# Public API
#--------------------------------------------------

#
# Generate a Rel query that will lookup the versions of Rel libraries currently installed in
# the database.
#
def generate_query():
    return _generate_query(DEPENDENCIES)

#
# Using the response from the query above, verify that the versions of the Rel libraries
# found in the database are compatible with the current dependencies of this python library.
#
# This function will raise an exception if dependencies are not met.
#
def check_dependencies(response:str, platform:str, config:cfg.Config, model_name:str):
    return _compare_dependencies(
        DEPENDENCIES,
        _extract_library_versions(response.results),
        _extract_std_version(response.results),
        platform,
        config,
        model_name
    )


#--------------------------------------------------
# Implementation Details
#--------------------------------------------------

def _extract_std_version(results):
    for result in results:
        if result["relationId"].startswith("/:output/:std"):
            for (version,) in result["table"].itertuples(index=False):
                return version
    return None

def _extract_library_versions(results):
    libraries = {}
    for result in results:
        if result["relationId"].startswith("/:output/:static_lock"):
            for (name, uuid, version) in result["table"].itertuples(index=False):
                libraries[(name, uuid)] = version
    return libraries

def _compare_dependencies(expected, lock, std, platform:str, config:cfg.Config, model_name:str):

    if platform == "":
        platform = "snowflake"

    # database is pre-versioning
    if not lock and not std:
        RAIInvalidVersionWarning(
            kind=InvalidVersionKind.SchemaOutOfDate,
            expected=expected,
            lock=lock,
            platform=platform,
            config=config,
            model_name=model_name,
        )
        return
    # database has std but no lock, it is using a package manager
    if not lock and std:
        RAIInvalidVersionWarning(
            kind=InvalidVersionKind.LibraryOutOfDate,
            expected=expected,
            lock=lock,
            platform=platform,
            config=config,
            model_name=model_name,
        )
        return

    errors = []
    for (name, uuid, range) in expected:
        key = (name, uuid)
        # library is missing from the database
        if key not in lock:
            errors.append((InvalidVersionKind.SchemaOutOfDate, name, uuid))
        else:
            version = Version(lock[key])
            # Rel library version higher than upper bound, too new
            if version >= range[1]:
                errors.append((InvalidVersionKind.LibraryOutOfDate, name, uuid))

            # Rel library version lower than lower bound, too old
            if version < range[0]:
                errors.append((InvalidVersionKind.SchemaOutOfDate, name, uuid))

    if errors:
        if all(kind == InvalidVersionKind.LibraryOutOfDate for (kind, _, _) in errors):
            kind = InvalidVersionKind.LibraryOutOfDate
        elif all(kind == InvalidVersionKind.SchemaOutOfDate for (kind, _, _) in errors):
            kind = InvalidVersionKind.SchemaOutOfDate
        else:
            kind = InvalidVersionKind.Incompatible
        RAIInvalidVersionWarning(
            kind=kind,
            expected=expected,
            lock=lock,
            platform=platform,
            errors=errors,
            config=config,
            model_name=model_name,
        )


def _generate_query(dependencies):
    bindings = " ;\n            ".join(
        ["(\"%s\", \"%s\")" % (name, uuid) for name, uuid, range in dependencies]
    )
    return f'''
    @no_diagnostics(:UNDEFINED_IDENTIFIER)
    def output[:std]: {{ std::version }}

    @no_diagnostics(:TYPE_MISMATCH)
    def output(:static_lock, name, uuid, version):
        rel(:pkg, :std, :pkg, :project, :static_lock, name, uuid, version) and
        {{
            {bindings}
        }}(name, uuid)'''
